/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.From;
import org.tranql.ql.Node;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.Select;
import org.tranql.ql.Where;
import org.tranql.query.QueryCommand;
import org.tranql.schema.Association;
import org.tranql.schema.Association.JoinDefinition;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.schema.FKAttribute;
import org.tranql.sql.EndTable;
import org.tranql.sql.SQLSchema;
import org.tranql.sql.Table;

/**
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
class AssociationEndFaultHandlerBuilder {
    private final SQLSchema schema;
    
    AssociationEndFaultHandlerBuilder(SQLSchema schema) {
        this.schema = schema;
    }
    
    public QueryCommand buildCMRFaultHandler(String name, String endName) throws QueryException {
        Table table = schema.getTable(name);
        EndTable end = (EndTable) table.getAssociationEnd(endName);
        if (end.isManyToMany()) {
            return buildManyToMany(end);
        } else if (end.isOnPKSide()) {
            return buildEndOnPKSide(end);
        } else {
            return buildEndOnFKSide(end);
        }
    }
    
    private QueryCommand buildManyToMany(EndTable end) throws QueryException {
        Association association = end.getAssociation();
        Entity mtmEntity = association.getManyToManyEntity();

        QuerySource t = new QuerySource(mtmEntity, "T");
        
        Select select = new Select(false);
        LinkedHashMap pkToFkMap;
        if ( association.getLeftJoinDefinition().getPKEntity() == end.getEntity() ) {
            pkToFkMap = association.getLeftJoinDefinition().getPKToFKMapping(); 
        } else {
            pkToFkMap = association.getRightJoinDefinition().getPKToFKMapping(); 
        }
        QueryBinding[] resultTransforms = new QueryBinding[pkToFkMap.size()];
        int index = 0;
        for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            FKAttribute att = (FKAttribute) entry.getValue();
            select.addChild(new FKAttributeReference(t, att));
            resultTransforms[index] = new QueryBindingImpl(index, mtmEntity, att);
            index++;
        }

        From from = new From();
        from.addChild(t);
        
        Where where = new Where();
        Node rootCond = null;
        if ( association.getLeftJoinDefinition().getPKEntity() == end.getEntity() ) {
            pkToFkMap = association.getRightJoinDefinition().getPKToFKMapping(); 
        } else {
            pkToFkMap = association.getLeftJoinDefinition().getPKToFKMapping(); 
        }
        QueryBinding[] paramTransforms = new QueryBinding[pkToFkMap.size()];
        index = 0;
        for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            FKAttribute att = (FKAttribute) entry.getValue();
            Node localCond = new BinaryOperation(BinaryOperation.EQ);
            localCond.addChild(new FKAttributeReference(t, att));
            localCond.addChild(new ParameterReference(index));
            rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
            paramTransforms[index] = new QueryBindingImpl(index, mtmEntity, att);
            index++;
        }
        where.addChild(rootCond);
        
        Query query = new Query(paramTransforms, resultTransforms);
        query.addChild(select).addChild(from).addChild(where);
    
        return schema.getCommandFactory().createQuery(query);
    }
    
    private QueryCommand buildEndOnPKSide(EndTable end) throws QueryException {
        Association association = end.getAssociation();
        JoinDefinition jDef = association.getJoinDefinition();
        Map pkToFkMap = jDef.getPKToFKMapping();
        Entity fkEntity = jDef.getFKEntity();

        QuerySource t = new QuerySource(fkEntity, "T");
        
        List resultTransforms = new ArrayList();
        Select select = new Select(false);
        List attributes = fkEntity.getPrimaryKeyFields();
        for (int i = 0; i < attributes.size(); i++) {
            Attribute att = (Attribute) attributes.get(i);
            select.addChild(new AttributeReference(t, att));
            resultTransforms.add(new QueryBindingImpl(resultTransforms.size(), fkEntity, att));
        }
        
        From from = new From();
        from.addChild(t);
        

        QueryBinding[] paramTransforms = new QueryBinding[pkToFkMap.size()];
        int index = 0;
        Where where = new Where();
        Node rootCond = null;
        for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            FKAttribute att = (FKAttribute) entry.getValue();
            Node localCond = new BinaryOperation(BinaryOperation.EQ);
            localCond.addChild(new FKAttributeReference(t, att));
            localCond.addChild(new ParameterReference(index));
            rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
            paramTransforms[index] = new QueryBindingImpl(index, fkEntity, att);
            index++;
        }
        where.addChild(rootCond);

        Query query = new Query(paramTransforms, (QueryBinding[]) resultTransforms.toArray(new QueryBinding[0]));
        query.addChild(select).addChild(from).addChild(where);
    
        return schema.getCommandFactory().createQuery(query);
    }
    
    private QueryCommand buildEndOnFKSide(EndTable end) throws QueryException {
        Association association = end.getAssociation();
        JoinDefinition jDef = association.getJoinDefinition();
        Map pkToFkMap = jDef.getPKToFKMapping();
        Entity fkEntity = jDef.getFKEntity();
        
        QuerySource t = new QuerySource(fkEntity, "T");
        
        Select select = new Select(false);
        QueryBinding[] resultTransforms = new QueryBinding[pkToFkMap.size()];
        int index = 0;
        for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            FKAttribute att = (FKAttribute) entry.getValue();
            select.addChild(new FKAttributeReference(t, att));
            resultTransforms[index] = new QueryBindingImpl(index, fkEntity, att);
            index++;
        }

        From from = new From();
        from.addChild(t);
    
        List paramTransforms = new ArrayList();
        Where where = new Where();
        List attributes = fkEntity.getPrimaryKeyFields();
        Node rootCond = null;
        index = 0;
        for (int i = 0; i < attributes.size(); i++) {
            Attribute att = (Attribute) attributes.get(i);
            Node localCond = new BinaryOperation(BinaryOperation.EQ);
            localCond.addChild(new AttributeReference(t, att));
            localCond.addChild(new ParameterReference(index));
            rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
            paramTransforms.add(new QueryBindingImpl(index, fkEntity, att));
            index++;
        }
        where.addChild(rootCond);
        
        Query query = new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), resultTransforms);
        query.addChild(select).addChild(from).addChild(where);
    
        return schema.getCommandFactory().createQuery(query);
    }
}
