/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import org.tranql.cache.*;
import org.tranql.ejb.*;
import org.tranql.field.*;
import org.tranql.identity.IdentityDefiner;
import org.tranql.identity.IdentityTransform;
import org.tranql.ql.*;
import org.tranql.query.QueryCommand;
import org.tranql.query.UpdateCommand;
import org.tranql.schema.*;
import org.tranql.schema.Association.JoinDefinition;

import java.util.*;

/**
 *
 *
 * @version $Revision: 431 $ $Date: 2006-11-13 18:18:35 -0500 (Mon, 13 Nov 2006) $
 */
public class EJBQueryBuilder {
    private final EJBSchema schema;
    private final IdentityDefinerBuilder identityDefinerBuilder;

    public EJBQueryBuilder(IdentityDefinerBuilder identityDefinerBuilder) {
        this.schema = identityDefinerBuilder.getEJBSchema();
        this.identityDefinerBuilder = identityDefinerBuilder;
    }

    /**
     * SELECT OBJECT(e) FROM ejb AS e WHERE e.pk = ?
     * @param name the global name of the EJB
     * @return a select Query
     */
    public QueryCommand buildFindByPrimaryKey(String name, boolean local) throws QueryException {
        EJB ejb = schema.getEJB(name);
        QuerySource source = new QuerySource(ejb, ejb.getName().substring(0, 1), ejb.isSelectForUpdate());

        List pkFields = ejb.getPrimaryKeyFields();
        QueryBinding[] resultTransforms = new QueryBinding[pkFields.size()];
        for (int i = 0; i < resultTransforms.length; i++) {
            Attribute attribute = (Attribute) pkFields.get(i);
            resultTransforms[i] = new QueryBindingImpl(i, ejb, attribute);
        }

        Node select = new Select(false);
        select.addChild(new EntityReference(source));
        Node from = new From().addChild(source);

        QueryBinding[] paramTransforms = new QueryBinding[pkFields.size()];
        
        IdentityDefiner identityDefiner = identityDefinerBuilder.getIdentityDefiner(ejb);
        IdentityTransform identityTransform = identityDefinerBuilder.getPrimaryKeyTransform(ejb);
        
        Where where = new Where();
        Node rootCond = null;
        for (int i = 0; i < paramTransforms.length; i++) {
            Attribute att = (Attribute) pkFields.get(i);
            Node localCond = new BinaryOperation(BinaryOperation.EQ);
            localCond.addChild(new AttributeReference(source, att));
            localCond.addChild(new ParameterReference(i));
            rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
            paramTransforms[i] = new QueryBindingImpl(new NestedRowAccessor(
                new IdentityExtractorAccessor(new GlobalIdentityAccessor(new FieldAccessor(0, null), identityTransform), identityDefiner),
                i,
                att.getType()), ejb, att);
        }
        where.addChild(rootCond);

        Query query = (Query) new Query(paramTransforms, resultTransforms).addChild(select).addChild(from).addChild(where);
        QueryCommand queryCommand = schema.getCommandFactory().createQuery(query);
        
        identityDefiner = identityDefinerBuilder.getIdentityDefiner(ejb, 0);
        EJBProxyFactory proxyFactory = ejb.getProxyFactory();
        FieldTransform[] view = new FieldTransform[1]; 
        view[0] = new ReferenceAccessor(identityDefiner);
        view[0] = new DomainIdentityAccessor(view[0], identityTransform);
        if ( local ) {
            view[0] = new EJBLocalObjectAsIdTransform(view[0], proxyFactory, ejb.getPrimaryKeyClass());
        } else {
            view[0] = new EJBObjectAsIdTransform(view[0], proxyFactory, ejb.getPrimaryKeyClass());
        }

        queryCommand.getQuery().setResultAccessors(view);
        
        return queryCommand;
    }

    /**
     * SELECT e.a1, e.a2 FROM ejb AS e WHERE e.pk = ?
     * @param name the global name of the EJB
     * @param attrNames the name of the attributes to load
     * @return a select Query
     */
    public QueryCommand buildLoad(String name, String[] attrNames) throws QueryException {
        EJB ejb = schema.getEJB(name);
        QuerySource source = new QuerySource(ejb, ejb.getName().substring(0, 1), ejb.isSelectForUpdate());
        Node from = new From().addChild(source);

        // output is listed fields
        Node select = new Select(false);
        QueryBinding[] resultTransforms = new QueryBinding[attrNames.length];
        FieldTransform[] resultAccessors = new FieldTransform[attrNames.length];
        for (int i=0; i < attrNames.length; i++) {
            Attribute attr = ejb.getAttribute(attrNames[i]);
            resultTransforms[i] = new QueryBindingImpl(i, ejb, attr);
            select.addChild(new AttributeReference(source, attr));
            resultAccessors[i] = new FieldAccessor(i, attr.getType());
        }

        List paramTransforms = new ArrayList();
        Node where = getWhere(source, paramTransforms, false);
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), resultTransforms).addChild(select).addChild(from).addChild(where);
        query.setResultAccessors(resultAccessors);
        return schema.getCommandFactory().createQuery(query);
    }

    /**
     * INSERT INTO ejb VALUES(...)
     * @param name the global name of the EJB
     * @return an insert query
     */
    public UpdateCommand buildCreate(String name) throws QueryException {
        EJB ejb = schema.getEJB(name);
        Insert insert = new Insert(ejb);

        List attributes = ejb.getAttributes();
        List paramTransforms = new ArrayList();
        int param = 0;
        
        RowConstructor insertList = new RowConstructor();
        RowConstructor values = new RowConstructor();

        for (int i=0; i < attributes.size(); i++) {
            Attribute attr = (Attribute) attributes.get(i);
            if (null != ejb.getAssociationEndDefiningFKAttribute(attr.getName())) {
                continue;
            }
            paramTransforms.add(new QueryBindingImpl(i, ejb, attr));
            insertList.addChild(new AttributeReference(insert, attr));
            values.addChild(new ParameterReference(param++));
        }

        List pkFields = ejb.getPrimaryKeyFields();
        int baseIdx = attributes.size();
        List associationEnds = ejb.getAssociationEnds();
        for (int i = 0; i < associationEnds.size(); i++) {
            AssociationEnd end = (AssociationEnd) associationEnds.get(i);
            IdentityDefiner identityDefiner = identityDefinerBuilder.getIdentityDefiner(end.getEntity());
            if (end.isOnPKSide()) {
                continue;
            }
            boolean detectSlotUnderConstruction = false;
            for (Iterator iterator = pkFields.iterator(); iterator.hasNext();) {
                Attribute pkField = (Attribute) iterator.next();
                if (end.hasFKAttribute(pkField.getName())) {
                    detectSlotUnderConstruction = true;
                    break;
                }
            }

            LinkedHashMap pkToFkMap = end.getAssociation().getJoinDefinition().getPKToFKMapping();
            int idIdx = 0;
            for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext(); idIdx++) {
                Map.Entry entry = (Map.Entry) iter.next();
                FKAttribute attribute = (FKAttribute) entry.getValue();
                FieldTransform accessor = new FieldAccessor(i + baseIdx, null);
                if (detectSlotUnderConstruction) {
                    accessor = new UnderConstructionSlotDetector(accessor, i + baseIdx);
                }

                paramTransforms.add(new QueryBindingImpl(
                    new NestedRowAccessor(
                        new IdentityExtractorAccessor(accessor, identityDefiner),
                            idIdx, attribute.getType()), ejb, attribute));
                insertList.addChild(new FKAttributeReference(insert, attribute));
                values.addChild(new ParameterReference(param++));
            }
        }
        
        insert.addChild(insertList).addChild(new TableConstructor().addChild(values));
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(insert);
        return schema.getCommandFactory().createUpdate(query);
    }

    /**
     * UPDATE ejb AS e SET e.f1 ?= ?, e.f2 ?= ? WHERE e.pk = ?
     * @param name
     * @return
     * @throws QueryException
     */
    public UpdateCommand buildStore(String name) throws QueryException {
        EJB ejb = schema.getEJB(name);
        List attributes = ejb.getAttributes();
        Update update = new Update(ejb);
        List paramTransforms = new ArrayList();

        SetList setList = new SetList();
        int slot = 0;
        int param = 0;
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            Attribute attr = (Attribute) i.next();
            if (!attr.isIdentity() && null == ejb.getAssociationEndDefiningFKAttribute(attr.getName())) {
                setList.addChild(new ConditionalAssignment().addChild(new AttributeReference(update, attr)).addChild(new ParameterReference(param++)).addChild(new ParameterReference(param++)));
                paramTransforms.add(new QueryBindingImpl(new ModifiedSlotDetector(slot)));
                paramTransforms.add(new QueryBindingImpl(new ModifiedSlotAccessor(slot, attr.getType()), ejb, attr));
            }
            slot++;
        }
        
        int baseIdx = attributes.size();
        List associationEnds = ejb.getAssociationEnds();
        for (int i = 0; i < associationEnds.size(); i++) {
            AssociationEnd end = (AssociationEnd) associationEnds.get(i);
            if (end.isOnPKSide()) {
                continue;
            }

            LinkedHashMap pkToFkMap = end.getAssociation().getJoinDefinition().getPKToFKMapping();
            int idIdx = 0;
            for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext(); idIdx++) {
                Map.Entry entry = (Map.Entry) iter.next();
                FKAttribute attribute = (FKAttribute) entry.getValue();

                Attribute att = ejb.getAttribute(attribute.getName());
                if (null != att && att.isIdentity()) {
                    continue;
                }
                
                setList.addChild(new ConditionalAssignment().addChild(new FKAttributeReference(update, attribute)).addChild(new ParameterReference(param++)).addChild(new ParameterReference(param++)));
                paramTransforms.add(new QueryBindingImpl(new ModifiedSlotDetector(i + baseIdx)));
                paramTransforms.add(new QueryBindingImpl(
                    new NestedRowAccessor(
                        new IdentityExtractorAccessor(new ModifiedSlotAccessor(i + baseIdx, null), identityDefinerBuilder.getIdentityDefiner(end.getEntity())),
                        idIdx, attribute.getType()), ejb, attribute));
            }
        }
        
        if ( null == setList.getChild() ) {
            return schema.getCommandFactory().createUpdate(NoOpQuery.NO_OP_QUERY);
        }
        
        update.addChild(setList).addChild(getWhere(update, paramTransforms, true));
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(update);
        return schema.getCommandFactory().createUpdate(query);
    }

    public UpdateCommand buildStore(String name, CacheRow cacheRow) throws QueryException {
        EJB ejb = schema.getEJB(name);
        List attributes = ejb.getAttributes();
        Update update = new Update(ejb);
        List paramTransforms = new ArrayList();

        SetList setList = new SetList();
        int slot = 0;
        int param = 0;
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            Attribute attr = (Attribute) i.next();
            if (!attr.isIdentity()
                    && null == ejb.getAssociationEndDefiningFKAttribute(attr.getName())
                    && cacheRow.isModified(slot)) {
                BinaryOperation equal = new BinaryOperation(BinaryOperation.EQ);
                equal.addChild(new AttributeReference(update, attr));
                equal.addChild(new ParameterReference(param++));
                setList.addChild(equal);
                
                paramTransforms.add(new QueryBindingImpl(new CacheRowAccessor(slot, attr.getType()), ejb, attr));
            }
            slot++;
        }
        
        int baseIdx = attributes.size();
        List associationEnds = ejb.getAssociationEnds();
        for (int i = 0; i < associationEnds.size(); i++) {
            if (false == cacheRow.isModified(i + baseIdx)) {
                continue;
            }
            
            AssociationEnd end = (AssociationEnd) associationEnds.get(i);
            if (end.isOnPKSide()) {
                continue;
            }

            LinkedHashMap pkToFkMap = end.getAssociation().getJoinDefinition().getPKToFKMapping();
            int idIdx = 0;
            for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext(); idIdx++) {
                Map.Entry entry = (Map.Entry) iter.next();
                FKAttribute attribute = (FKAttribute) entry.getValue();

                Attribute att = ejb.getAttribute(attribute.getName());
                if (null != att && att.isIdentity()) {
                    continue;
                }
                
                BinaryOperation equal = new BinaryOperation(BinaryOperation.EQ);
                equal.addChild(new FKAttributeReference(update, attribute));
                equal.addChild(new ParameterReference(param++));
                setList.addChild(equal);
                
                paramTransforms.add(new QueryBindingImpl(
                    new NestedRowAccessor(
                        new IdentityExtractorAccessor(new CacheRowAccessor(i + baseIdx, null), identityDefinerBuilder.getIdentityDefiner(end.getEntity())),
                        idIdx, attribute.getType()), ejb, attribute));
            }
        }
        
        if ( null == setList.getChild() ) {
            return schema.getCommandFactory().createUpdate(NoOpQuery.NO_OP_QUERY);
        }
        
        update.addChild(setList).addChild(getWhere(update, paramTransforms, true));
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(update);
        return schema.getCommandFactory().createUpdate(query);
    }
    
    public UpdateCommand buildRemove(String name) throws QueryException {
        EJB ejb = schema.getEJB(name);
        Delete delete = new Delete(ejb);

        List paramTransforms = new ArrayList();
        delete.addChild(getWhere(delete, paramTransforms, true));
        
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(delete);
        return schema.getCommandFactory().createUpdate(query);
    }

    public UpdateCommand buildMTMCreate(Association association) throws QueryException {
        Entity mtmEntity = association.getManyToManyEntity();
        JoinDefinition[] joinDefinitions = new JoinDefinition[] {
            association.getLeftJoinDefinition(),
            association.getRightJoinDefinition()
        };
        
        Insert insert = new Insert(mtmEntity);

        List paramTransforms = new ArrayList();
        int param = 0;
        RowConstructor insertList = new RowConstructor();
        RowConstructor values = new RowConstructor();
        for (int i=0; i < 2; i++) {
            LinkedHashMap pkToFkMap = joinDefinitions[i].getPKToFKMapping();
            int idIdx = 0;
            for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                FKAttribute attribute = (FKAttribute) entry.getValue();
                paramTransforms.add(new QueryBindingImpl(
                    new NestedRowAccessor(
                        new IdentityExtractorAccessor(new FieldAccessor(i, null), identityDefinerBuilder.getIdentityDefiner(joinDefinitions[i].getPKEntity())),
                        idIdx++, attribute.getType()), mtmEntity, attribute));
                insertList.addChild(new FKAttributeReference(insert, attribute));
                values.addChild(new ParameterReference(param++));
            }
        }
        
        insert.addChild(insertList).addChild(new TableConstructor().addChild(values));
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(insert);
        return schema.getCommandFactory().createUpdate(query);
    }

    public UpdateCommand buildMTMRemove(Association association) throws QueryException {
        Entity mtmEntity = association.getManyToManyEntity();
        JoinDefinition[] joinDefinitions = new JoinDefinition[] {
            association.getLeftJoinDefinition(),
            association.getRightJoinDefinition()
        };
        
        Delete delete = new Delete(mtmEntity);

        List paramTransforms = new ArrayList();
        Node rootCond = null;
        for (int i = 0; i < 2; i++) {
            LinkedHashMap pkToFkMap = joinDefinitions[i].getPKToFKMapping();
            int idIdx = 0;
            for (Iterator iter = pkToFkMap.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                FKAttribute attribute = (FKAttribute) entry.getValue();
                Node localCond = new BinaryOperation(BinaryOperation.EQ);
                localCond.addChild(new FKAttributeReference(delete, attribute));
                localCond.addChild(new ParameterReference(paramTransforms.size()));
                rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
                paramTransforms.add(new QueryBindingImpl(
                    new NestedRowAccessor(
                        new IdentityExtractorAccessor(new FieldAccessor(i, null), identityDefinerBuilder.getIdentityDefiner(joinDefinitions[i].getPKEntity())),
                        idIdx++, attribute.getType()), mtmEntity, attribute));
            }
        }

        delete.addChild(new Where().addChild(rootCond));
        
        Query query = (Query) new Query((QueryBinding[]) paramTransforms.toArray(new QueryBinding[0]), new QueryBinding[0]).addChild(delete);
        return schema.getCommandFactory().createUpdate(query);
    }
    
    private Node getWhere(AliasedEntity source, List paramTransforms, boolean isParamTableCacheRow) {
        EJB ejb = (EJB) source.getEntity();
        List pkFields = ejb.getPrimaryKeyFields();
        List association = ejb.getAttributes();
        int baseParamIdx = paramTransforms.size();
        
        Where where = new Where();
        Node rootCond = null;
        for (int i = 0; i < pkFields.size(); i++) {
            Attribute att = (Attribute) pkFields.get(i);
            Node localCond = new BinaryOperation(BinaryOperation.EQ);
            localCond.addChild(new AttributeReference(source, att));
            localCond.addChild(new ParameterReference(baseParamIdx + i));
            rootCond = null == rootCond ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(rootCond).addChild(localCond);
            int paramIdx = isParamTableCacheRow ? association.indexOf(att) : i;
            paramTransforms.add(new QueryBindingImpl(paramIdx, ejb, att));
        }
        where.addChild(rootCond);

        return where;
    }
}
