/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.ejb.EJBLocalObject;
import javax.ejb.EJBObject;

import org.tranql.ejb.IdAsEJBLocalObjectTransform;
import org.tranql.ejb.IdAsEJBObjectTransform;
import org.tranql.field.GlobalIdentityAccessor;
import org.tranql.field.IdentityExtractorAccessor;
import org.tranql.field.NestedRowAccessor;
import org.tranql.identity.IdentityDefiner;
import org.tranql.identity.IdentityTransform;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.Node;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Path;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryBindingImpl;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QueryVisitor;
import org.tranql.ql.Where;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.schema.FKAttribute;


/**
 * 
 * 
 * @version $Revision: 410 $ $Date: 2006-10-21 20:15:17 -0400 (Sat, 21 Oct 2006) $
 */
public class QueryTileBuilder {

    public static LinkedHashMap buildQueryBindings(QueryBinding[] bindings, int slot, Entity entity, IdentityDefinerBuilder identityDefinerBuilder) throws QueryException {
        IdentityDefiner identityDefiner = identityDefinerBuilder.getIdentityDefiner(entity);
        IdentityTransform identityTransform = identityDefinerBuilder.getPrimaryKeyTransform(entity);
        
        QueryBinding binding = bindings[slot];
        Class clazz = binding.getFieldClass();
       if (EJBLocalObject.class.isAssignableFrom(clazz)) {
            binding = new QueryBindingImpl(new IdAsEJBLocalObjectTransform(binding, null, null));
        } else if (EJBObject.class.isAssignableFrom(clazz)) {
            binding = new QueryBindingImpl(new IdAsEJBObjectTransform(binding, null, null));
        } else {
            throw new QueryException(binding + " does not correspond to an EJBLocalObject or EJBObject parameter.");
        }
        
        LinkedHashMap newBindings = new LinkedHashMap();
        List pkKeys = entity.getPrimaryKeyFields();
        for (int i = 0; i < pkKeys.size(); i++) {
            Attribute attribute = (Attribute) pkKeys.get(i);
            ParameterReference newParameter = new ParameterReference(bindings.length + i);
            QueryBinding newBinding = new QueryBindingImpl(
                    new NestedRowAccessor(
                            new IdentityExtractorAccessor(
                                    new GlobalIdentityAccessor(
                                            binding, identityTransform), 
                                            identityDefiner), i, null), entity, attribute);
            newBindings.put(newParameter, newBinding);
        }
        
        return newBindings;
    }
    
    public static Where buildWhere(Where where, Node appended, QueryVisitor visitor, Object param) throws QueryException {
        Where to = new Where();
        
        if (null != appended) {
            Node childNode = where.getChild();
            if (null != childNode) {
                appended = new BinaryOperation(BinaryOperation.AND).addChild(appended); 
            }
            to.addChild(appended);
            visitChildren(appended, where, visitor, param);
        } else {
            visitChildren(to, where, visitor, param);
        }

        return to;
    }
    
    public static Node buildJoinCondition(Collection factories) throws QueryException {
        Node returnedNode = null;
        
        Set processed = new HashSet();
        for (Iterator iter = factories.iterator(); iter.hasNext();) {
            QuerySourceFactory factory = (QuerySourceFactory) iter.next();
            Node localCond = buildJoinCondition(factory, processed);
            if (null == localCond) {
                continue;
            }
            returnedNode = null == returnedNode ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(returnedNode).addChild(localCond);
        }
        
        return returnedNode;
    }
    
    public static Node buildJoinCondition(Association.JoinDefinition joinDefinition, AliasedEntity qs1, AliasedEntity qs2) throws QueryException {
        return buildJoinCondition(joinDefinition, qs1, qs2, new HashSet());
    }

    
    public static Node buildJoinCondition(QuerySourceFactory factory) throws QueryException {
        return buildJoinCondition(factory, new HashSet());
    }

    private static void visitChildren(Node to, Node from, QueryVisitor visitor, Object param) throws QueryException {
        for (Node fromChild = from.getChild(); fromChild != null; fromChild = fromChild.getSibling()) {
            Node toChild = (Node) fromChild.visit(visitor, param);
            if (toChild != null) {
                to.addChild(toChild);
            }
        }
    }    
    
    private static Node buildJoinCondition(QuerySourceFactory factory, Set processed) throws QueryException {
        Node returnedNode = null;
        
        List thePath = factory.getPath().getPath();
        QuerySource prevQS = factory.factory(0);
        for (int i = 0; i < thePath.size(); i++) {
            Object opaque = thePath.get(i);
            if (false == opaque instanceof AssociationEnd) {
                break;
            }
            AssociationEnd end = (AssociationEnd) opaque;
            QuerySource qs = factory.factory(i + 1);

            Node localCond;
            
            Association association = end.getAssociation();
            if (end.isManyToMany()) {
                Association.JoinDefinition joinDefinition = association.getLeftJoinDefinition();
                QuerySource mtmQS = factory.factoryMTM();
                
                Node leftCond;
                if (joinDefinition.getPKEntity() == prevQS.getEntity()) {
                    leftCond = buildJoinCondition(joinDefinition, prevQS, mtmQS, processed);
                } else {
                    leftCond = buildJoinCondition(joinDefinition, qs, mtmQS, processed);
                }
                
                joinDefinition = association.getRightJoinDefinition();

                Node rightCond;
                if (joinDefinition.getPKEntity() == prevQS.getEntity()) {
                    rightCond = buildJoinCondition(joinDefinition, prevQS, mtmQS, processed);
                } else {
                    rightCond = buildJoinCondition(joinDefinition, qs, mtmQS, processed);
                }
                
                localCond = new BinaryOperation(BinaryOperation.AND).addChild(leftCond).addChild(rightCond);
            } else {
                localCond = buildJoinCondition(association.getJoinDefinition(), prevQS, qs, processed);
            }

            returnedNode = null == returnedNode ? localCond : new BinaryOperation(BinaryOperation.AND).addChild(returnedNode).addChild(localCond);

            prevQS = qs;
        }
        
        return returnedNode;
    }
    
    private static Node buildJoinCondition(Association.JoinDefinition joinDefinition, AliasedEntity qs1, AliasedEntity qs2, Set joined) throws QueryException {
        AliasedEntity pkQS;
        AliasedEntity fkQS;
        if (qs1.getEntity() == joinDefinition.getPKEntity() && qs2.getEntity() == joinDefinition.getFKEntity()) {
            pkQS = qs1;
            fkQS = qs2;
        } else if (qs2.getEntity() == joinDefinition.getPKEntity() && qs1.getEntity() == joinDefinition.getFKEntity()) {
            pkQS = qs2;
            fkQS = qs1;
        } else {
            throw new AssertionError();
        }

        if (false == joined.add(pkQS.getAlias() + "." + fkQS.getAlias())) {
            return null;
        }
        
        Node returnedNode = null;
        
        LinkedHashMap pkToFkMapping = joinDefinition.getPKToFKMapping();
        for (Iterator iterator = pkToFkMapping.entrySet().iterator(); iterator.hasNext();) {
            Map.Entry element = (Map.Entry) iterator.next();
            Attribute attribute = (Attribute) element.getKey();
            FKAttribute fkAttribute = (FKAttribute) element.getValue();
            Node localCondition = new BinaryOperation(BinaryOperation.EQ);
            localCondition.addChild(new AttributeReference(pkQS, attribute));
            localCondition.addChild(new FKAttributeReference(fkQS, fkAttribute));
            returnedNode = null == returnedNode ? localCondition : new BinaryOperation(BinaryOperation.AND).addChild(returnedNode).addChild(localCondition);
        }
        
        return returnedNode;
    }

    public interface QuerySourceFactory {
        QuerySource factory(int index) throws QueryException;

        QuerySource factoryMTM() throws QueryException;

        Path getPath();
    }

}
