/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.builder;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.tranql.cache.GlobalSchema;
import org.tranql.ejb.CMPField;
import org.tranql.ejb.CMRField;
import org.tranql.ejb.EJB;
import org.tranql.ejb.EJBSchema;
import org.tranql.ejb.FinderEJBQLQuery;
import org.tranql.ejb.SelectEJBQLQuery;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.EntityReference;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.From;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.Where;
import org.tranql.query.QueryCloner;
import org.tranql.query.QueryCommand;
import org.tranql.query.SchemaMapper;
import org.tranql.schema.Association;
import org.tranql.schema.Association.JoinDefinition;
import org.tranql.sql.EJBQLToPhysicalQuery;
import org.tranql.sql.EndTable;
import org.tranql.sql.SQLSchema;
import org.tranql.sql.Table;
import org.tranql.sql.prefetch.PrefetchGroup;
import org.tranql.sql.prefetch.PrefetchGroupDictionary;
import org.tranql.sql.prefetch.PrefetchGroupTransform;

/**
 *
 *
 * @version $Revision: 459 $ $Date: 2006-12-13 06:04:42 -0500 (Wed, 13 Dec 2006) $
 */
public class SQLQueryBuilder {
    private static final Log log = LogFactory.getLog(SQLQueryBuilder.class);
    
    private final EJBSchema ejbSchema;
    private final SQLSchema sqlSchema;
    private final GlobalSchema cacheSchema;
    private final SchemaMapper mapper;
    private final EJBQueryBuilder ejbQueryBuilder;
    private final EJBQLToPhysicalQuery ejbQLQueryBuilder;
    private final AssociationEndFaultHandlerBuilder endFaultHandlerBuilder;
    private final IdentityDefinerBuilder identityDefinerBuilder;
    private final PrefetchGroupDictionary groupDictionary;

    public SQLQueryBuilder(EJBSchema ejbSchema, SQLSchema sqlSchema, GlobalSchema cacheSchema) {
        this.ejbSchema = ejbSchema;
        this.sqlSchema = sqlSchema;
        this.cacheSchema = cacheSchema;
        mapper = new SchemaMapper(sqlSchema);
        ejbQueryBuilder = new EJBQueryBuilder(new IdentityDefinerBuilder(ejbSchema, cacheSchema));
        ejbQLQueryBuilder = new EJBQLToPhysicalQuery(ejbSchema, sqlSchema, cacheSchema);
        endFaultHandlerBuilder = new AssociationEndFaultHandlerBuilder(sqlSchema);
        identityDefinerBuilder = new IdentityDefinerBuilder(ejbSchema, cacheSchema);
        groupDictionary = sqlSchema.getGroupDictionary();
    }
    
    public QueryCommand buildFindByPrimaryKey(String name, boolean local) throws QueryException {
        QueryCommand command = ejbQueryBuilder.buildFindByPrimaryKey(name, local);
        command = mapper.transform(command);
        EJB ejb = ejbSchema.getEJB(name);
        Class pkClass = ejb.isUnknownPK() ? Object.class :  ejb.getPrimaryKeyClass();
        FinderEJBQLQuery pkFinder = new FinderEJBQLQuery("findByPrimaryKey", new Class[] {pkClass}, null);
        Collection finders = ejb.getFinders();
        boolean transformed = false;
        AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
        for (Iterator iter = finders.iterator(); iter.hasNext();) {
            FinderEJBQLQuery finder = (FinderEJBQLQuery) iter.next();
            if (finder.equals(pkFinder)) {
                if (null != finder.getPrefetchGroup()) {
                    String groupName = finder.getPrefetchGroup();
                    PrefetchGroup group = groupDictionary.getPrefetchGroup(groupName, name);
                    PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                            sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                            identityDefinerBuilder);
                    command = groupTransform.transform(command);
                    transformed = true;
                }
                break;
            }
        }
        if (!transformed) {
            PrefetchGroup group = groupDictionary.getDefaultPrefetchGroup(name);
            PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                    sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                    identityDefinerBuilder);
            command = groupTransform.transform(command);
        }
        
        log.debug("CMP [" + ejb + "] - [" + pkFinder + "] compiled to [" + command + "]");
        
        return command;
    }

    public QueryCommand buildLoadEntity(String name, String[] attributes) throws QueryException {
        QueryCommand command = ejbQueryBuilder.buildLoad(name, attributes);
        command = mapper.transform(command);

        EJB ejb = ejbSchema.getEJB(name);
        String groupName = ejb.getPrefetchGroup();
        PrefetchGroup group;
        if (null == groupName) {
            group = groupDictionary.getDefaultPrefetchGroup(name);
        } else {
            group = groupDictionary.getPrefetchGroup(groupName, name);
        }
        AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
        PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                identityDefinerBuilder);
        command = groupTransform.transform(command);

        log.debug("CMP [" + ejb + "] - entity load command compiled to [" + command + "]");

        return command;
    }

    public QueryCommand buildLoadAttribute(String name, String attrName, boolean prefetch) throws QueryException {
        QueryCommand command = ejbQueryBuilder.buildLoad(name, new String[] {attrName});
        command = mapper.transform(command);
        
        if (false == prefetch) {
            return command;
        }

        EJB ejb = ejbSchema.getEJB(name);
        CMPField field = (CMPField) ejb.getAttribute(attrName);
        String groupName = field.getPrefetchGroup();
        PrefetchGroup group;
        if (null != groupName) {
            group = groupDictionary.getPrefetchGroup(groupName, name);
        } else {
            group = groupDictionary.getGroupByAttributeName(name, attrName);
            if (null == group) {
                group = groupDictionary.getDefaultPrefetchGroup(name);
            }
        }
        
        AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
        PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                identityDefinerBuilder);
        command = groupTransform.transform(command);

        log.debug("CMP [" + ejb + "] - CMP Field load command for [" + field
            + "] compiled to [" + command + "]");

        return command;
    }

    public Map buildFinders(String ejbName) throws QueryException {
        EJB ejb = ejbSchema.getEJB(ejbName);
        Map queries = ejbQLQueryBuilder.buildFinders(ejb);
        for (Iterator iter = queries.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            FinderEJBQLQuery query = (FinderEJBQLQuery) entry.getKey();
            QueryCommand[] commands = (QueryCommand[]) entry.getValue();
            for (int i = 0; i < commands.length; i++) {
                QueryCommand command = commands[i];
                PrefetchGroup group;
                if (null != query.getPrefetchGroup()) {
                    String groupName = query.getPrefetchGroup();
                    group = groupDictionary.getPrefetchGroup(groupName, ejb.getName());
                } else {
                    group = groupDictionary.getDefaultPrefetchGroup(ejb.getName());
                }
                AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
                PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                        sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                        identityDefinerBuilder);
                command = groupTransform.transform(command);
                commands[i] = command;
            }

            log.debug("CMP [" + ejb + "] - [" + query + "] compiled to [" + commands[0] + "]");

            entry.setValue(commands);
        }
        return queries;
    }
    
    public Map buildSelects(String ejbName) throws QueryException {
        EJB ejb = ejbSchema.getEJB(ejbName);
        Map queries = ejbQLQueryBuilder.buildSelects(ejb);
        for (Iterator iter = queries.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            SelectEJBQLQuery query = (SelectEJBQLQuery) entry.getKey();
            QueryCommand command = (QueryCommand) entry.getValue();
            if (null != query.getSelectedEJB()){
                EJB selectedEJB = query.getSelectedEJB();
                PrefetchGroup group;
                if (null != query.getPrefetchGroup()) {
                    String groupName = query.getPrefetchGroup();
                    group = groupDictionary.getPrefetchGroup(groupName, selectedEJB.getName());
                } else {
                    group = groupDictionary.getDefaultPrefetchGroup(selectedEJB.getName());
                }
                AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
                PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                        sqlSchema, aliasedEntity, group, new QuerySourceDictionary(),
                        identityDefinerBuilder);
                command = groupTransform.transform(command);
            }

            log.debug("CMP [" + ejb + "] - [" + query + "] compiled to [" + command + "]");

            entry.setValue(command);
        }
        return queries;
    }
    
    public QueryCommand buildLoadAssociationEnd(String ejbName, String endName, boolean prefetch) throws QueryException {
        EJB ejb = ejbSchema.getEJB(ejbName);
        CMRField cmrField = (CMRField) ejb.getAssociationEnd(endName);
        Table table = sqlSchema.getTable(ejbName);
        EndTable endTable = (EndTable) table.getAssociationEnd(endName);
        QueryCommand command = endFaultHandlerBuilder.buildCMRFaultHandler(ejbName, endName);
        
        if (false == prefetch) {
            return command;
        }
        
        String groupName = cmrField.getPrefetchGroup();
        PrefetchGroup group;
        if (null != groupName) {
            group = groupDictionary.getPrefetchGroup(groupName, ejbName);
        } else {
            group = groupDictionary.getGroupByEndName(ejbName, endName);
            if (null == group) {
                group = groupDictionary.getDefaultGroupByEndName(ejbName, endName);
            }
        }
        
        AliasedEntity aliasedEntity = getSelectedAliasEntity(command);
        Association association = endTable.getAssociation();
        QuerySourceDictionary sourceDictionary = new QuerySourceDictionary();
        if (endTable.isManyToMany()) {
            JoinDefinition joinDefinition = association.getLeftJoinDefinition();
            String alias = sourceDictionary.getNextAlias();
            QuerySource toAppend = new QuerySource(joinDefinition.getPKEntity(), alias, ejb.isSelectForUpdate());
            if (table != toAppend.getEntity()) {
                joinDefinition = association.getRightJoinDefinition();
                toAppend = new QuerySource(joinDefinition.getPKEntity(), alias, ejb.isSelectForUpdate());
            }
            QuerySourceAppender sourceAppender = new QuerySourceAppender(joinDefinition, aliasedEntity, toAppend);
            command = sourceAppender.transform(command);
            aliasedEntity = toAppend;
        } else if (endTable.isOnPKSide()) {
            JoinDefinition joinDefinition = association.getJoinDefinition();
            String alias = sourceDictionary.getNextAlias();
            QuerySource toAppend = new QuerySource(table, alias, ejb.isSelectForUpdate());
            QuerySourceAppender sourceAppender = new QuerySourceAppender(joinDefinition, aliasedEntity, toAppend);
            command = sourceAppender.transform(command);
            aliasedEntity = toAppend;
        }
        PrefetchGroupTransform groupTransform = new PrefetchGroupTransform(
                sqlSchema, aliasedEntity, group, sourceDictionary, identityDefinerBuilder);
        command = groupTransform.transform(command);

        log.debug("CMP [" + ejb + "] - CMR Field load command for [" + cmrField
            + "] compiled to [" + command + "]");

        return command;
    }
    
    private AliasedEntity getSelectedAliasEntity(QueryCommand command) {
        Node node = command.getQuery().getChild().getChild();
        if (node instanceof AttributeReference) {
            AttributeReference reference = (AttributeReference) node;
            return reference.getSource();
        } else if (node instanceof FKAttributeReference) {
            FKAttributeReference reference = (FKAttributeReference) node;
            return reference.getSource();
        } else if (node instanceof EntityReference) {
            EntityReference reference = (EntityReference) node;
            return reference.getEntity();
        }

        throw new AssertionError("Unsupported selected node " + node);
    }
    
    private class QuerySourceAppender extends QueryCloner {
        private final JoinDefinition joinDefinition;
        private final QuerySource relatedQS;
        private final AliasedEntity sourceQS;
        
        private QuerySourceAppender(JoinDefinition joinDefinition, AliasedEntity sourceQS, QuerySource relatedQS) {
            this.joinDefinition = joinDefinition;
            this.sourceQS = sourceQS;
            this.relatedQS = relatedQS;
        }
        
        public QueryCommand transform(QueryCommand source) throws QueryException {
            return sqlSchema.getCommandFactory().createQuery((Query) super.visit(source.getQuery(), null));
        }
        
        public Object visit(From node, Object param) throws QueryException {
            From to = (From) super.visit(node, param);
            
            to = new From();
            to.addChild(new Join(Join.CROSS_JOIN).addChild(sourceQS).addChild(relatedQS));
            return to;
        };
        
        public Object visit(Where node, Object param) throws QueryException {
            Where to = (Where) super.visit(node, param);
            Node whereClause = to.getChild();
            Node condition = QueryTileBuilder.buildJoinCondition(joinDefinition, sourceQS, relatedQS);
            
            to = new Where();
            to.addChild(new BinaryOperation(BinaryOperation.AND).addChild(whereClause).addChild(condition));
            return to;
        };
    }
}
