/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.cache;

import java.util.BitSet;

import org.tranql.identity.GlobalIdentity;
import org.tranql.field.Row;

/**
 * The representation of a data set in the cache.
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public final class CacheRow extends Row {
    /**
     * A place holder indicating that the specified slot in the cache line
     * has not been loaded. Used to differentiate between NULL and missing values.
     */
    public static final Object NO_DATA = new Object() {
        public final String toString() {
            return "NO_DATA";
        }
    };

    /**
     * A place holder indicating that the specified slot in the cache line
     * should contain a Set to hold references to other rows.
     */
    public static final Object SET_VALUE = new Object() {
        public final String toString() {
            return "SET_VALUE";
        }
    };

    private final GlobalIdentity id;
    private final BitSet modified;
    private final BitSet underConstruction;
    private CacheRowState state;

    /**
     * Constructor used during creation of an instance.
     * @param defaults default values used to initialize each slot
     */
    public CacheRow(Object[] defaults) {
        super(new Object[defaults.length]);
        System.arraycopy(defaults, 0, values, 0, defaults.length);
        modified = new BitSet(defaults.length);
        underConstruction = new BitSet(defaults.length);
        this.id = null;
        state = CacheRowState.UNDER_CONSTRUCTION;
    }

    CacheRow(GlobalIdentity id, CacheRow source) {
        super(source.values);
        this.id = id;
        this.state = CacheRowState.NEW;
        this.modified = source.modified;
        this.underConstruction = source.underConstruction;
    }

    CacheRow(GlobalIdentity id, Object[] values) {
        super(values);
        this.id = id;
        this.state = CacheRowState.CLEAN;
        modified = new BitSet(values.length);
        underConstruction = new BitSet(values.length);
    }

    public void mergeWithRemoved(CacheRow deletedRow) {
        if (state != CacheRowState.NEW ||
                deletedRow.state != CacheRowState.REMOVED) {
            throw new IllegalArgumentException();
        }
        state = CacheRowState.MODIFIED;
        for (int i = 0; i < values.length; i++) {
            if (isModified(i)) {
                continue;
            }
            markDirty(i);
        }
    }

    
    /**
     * Return the state if this row.
     * @return the state of this row
     */
    public CacheRowState getState() {
        return state;
    }

    /**
     * Test if a slot has been modified.
     * @param slot the slot to check
     * @return true if the slot has been modified
     */
    public boolean isModified(int slot) {
        return modified.get(slot);
    }

    public boolean isLoaded(int slot) {
        return values[slot] != NO_DATA;
    }

    /**
     * Indicate that the state has been synchronized.
     */
    public void markClean() {
        modified.clear();
        state = CacheRowState.CLEAN;
    }

    public void markDirty(int slot) {
        if (state == CacheRowState.CLEAN) {
            state = CacheRowState.MODIFIED;
        }
        modified.set(slot);
    }

    public void markRemoved() {
        state = CacheRowState.REMOVED;
    }

    public void markSlotUnderConstruction(int slot) {
        if (state != CacheRowState.UNDER_CONSTRUCTION) {
            throw new IllegalStateException("Row is not under construction");
        }
        underConstruction.set(slot);
    }

    public void unmarkSlotUnderConstruction(int slot) {
        if (false == underConstruction.get(slot)) {
            throw new IllegalStateException("Slot is not under construction");
        }
        underConstruction.clear(slot);
    }

    public boolean isSlotUnderConstruction(int slot) {
        return underConstruction.get(slot);
    }

    /**
     * Return the identity of this row.
     * @return the global identity of this row
     */
    public GlobalIdentity getId() {
        if (this.id == null) {
            throw new IllegalStateException("Identity not available");
        }
        return id;
    }
}
