/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.cache;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.tranql.ql.QueryException;
import org.tranql.query.UpdateCommand;
import org.tranql.schema.EntityComparator;

/**
 * This flushing strategy enforces the relationships between the CacheTable
 * instances to be flushed; it enforces the foreign key constraints that may 
 * exist at the database level.
 *
 * @version $Revision: 276 $ $Date: 2005-10-01 04:41:03 -0400 (Sat, 01 Oct 2005) $
 */
public class EnforceRelationshipsFlushStrategy implements CacheFlushStrategy {
    public void rowAssociated(CacheRow row) {
    }
    
    public void rowAdded(CacheRow row) {
    }

    public void rowModified(CacheRow row) {
    }

    public void rowRemoved(CacheRow row) {
    }

    public void flush(Map cache) throws QueryException {
        // order the CacheTable instances to be flushed.
        Set tables = cache.keySet();
        TreeSet orderedTables = new TreeSet(new EntityComparator());
        orderedTables.addAll(tables);

        // get the UpdateCommand to be flushed.
        List inserts = new ArrayList();
        List updates = new ArrayList();
        List deletes = new ArrayList();
        for (Iterator iter = orderedTables.iterator(); iter.hasNext();) {
            CacheTable table = (CacheTable) iter.next();
            UpdateCommand insertCommand = table.getInsertCommand();
            UpdateCommand updateCommand = table.getUpdateCommand();
            UpdateCommand deleteCommand = table.getDeleteCommand();
            Map tableMap = (Map) cache.get(table);
            for (Iterator rows = tableMap.values().iterator(); rows.hasNext();) {
                CacheRow row = (CacheRow) rows.next();
                CacheRowState state = row.getState();
                if (state == CacheRowState.NEW) {
                    inserts.add(new UpdateCommandInfo(insertCommand, row));
                } else if (state == CacheRowState.MODIFIED) {
                    updates.add(new UpdateCommandInfo(updateCommand, row));
                } else if (state == CacheRowState.REMOVED) {
                    deletes.add(new UpdateCommandInfo(deleteCommand, row));
                }
            }
        }
        // delete operations are executed in the reverse order than the one
        // defined by the CacheTable instances.
        Collections.reverse(deletes);
        
        execute(inserts);
        // Implementation note: update operations can be executed in any order as
        // long as they are executed after the insert ones.
        execute(updates);
        execute(deletes);
    }

    private void execute(List updateCommandInfoList) throws QueryException {
        for (int i = 0; i < updateCommandInfoList.size(); i++) {
            UpdateCommandInfo updateCommandInfo = (UpdateCommandInfo) updateCommandInfoList.get(i);
            updateCommandInfo.execute();
        }
    }

    private static class UpdateCommandInfo {
        private final UpdateCommand command;
        private final CacheRow row;
        private UpdateCommandInfo(UpdateCommand command, CacheRow row) {
            this.command = command;
            this.row = row;
        }
        public void execute() throws QueryException {
            command.execute(null, row);
            row.markClean();
        }
    }
    
}
