/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejb;

import java.lang.reflect.Field;
import java.util.List;

import org.tranql.identity.GlobalIdentity;
import org.tranql.identity.IdentityTransform;
import org.tranql.identity.IdentityTransformException;
import org.tranql.identity.DerivedIdentityValue;
import org.tranql.cache.CacheTable;

/**
 * Identity transform that will return an new instance of the EJB primary key
 * for a cache row. The PK instance is instantiated and initialized using reflection.
 *
 * @version $Revision: 127 $ $Date: 2004-10-14 21:52:57 -0400 (Thu, 14 Oct 2004) $
 */
public class CompoundPKTransform implements IdentityTransform {
    private final CacheTable table;
    private final Class pkClass;
    private final List fieldNames;

    public CompoundPKTransform(CacheTable table, Class pkClass, List fieldNames) {
        this.table = table;
        this.pkClass = pkClass;
        this.fieldNames = fieldNames;
        Field[] fields = pkClass.getFields();
        for (int i = 0; i < fields.length; i++) {
            if ( false == fieldNames.contains(fields[i].getName()) ) {
                throw new IllegalArgumentException("Order of field " + fields[i].getName() + " of pkClass " + pkClass + " is not defined.");
            }
        }
    }

    public GlobalIdentity getGlobalIdentity(Object domainId) throws IdentityTransformException {
        if (!pkClass.isInstance(domainId)) {
            throw new IdentityTransformException("Id is not an instance of pkClass " + pkClass.getName());
        }
        Field[] fields = pkClass.getFields();
        Object[] values = new Object[fields.length];
        for (int i = 0; i < fields.length; i++) {
            try {
                values[fieldNames.indexOf(fields[i].getName())] = fields[i].get(domainId);
            } catch (IllegalArgumentException e) {
                throw new IdentityTransformException("Unable to convert value for field" + fields[i].getName(), e);
            } catch (IllegalAccessException e) {
                throw new IdentityTransformException("Unable to access field " + fields[i].getName(), e);
            }
        }
        return new GlobalIdentity(table, new DerivedIdentityValue(values));
    }

    public Object getDomainIdentity(GlobalIdentity globalId) throws IdentityTransformException {
        Object[] values = (Object[]) ((DerivedIdentityValue) globalId.getId()).getValues();
        Object pk;
        try {
            pk = pkClass.newInstance();
        } catch (InstantiationException e) {
            throw new IdentityTransformException("Unable to instantiate " + pkClass.getName(), e);
        } catch (IllegalAccessException e) {
            throw new IdentityTransformException("Unable to instantiate " + pkClass.getName(), e);
        } catch (IllegalArgumentException e) {
            throw new IdentityTransformException("Unable to instantiate " + pkClass.getName(), e);
        }

        Field[] fields = pkClass.getFields();
        for (int i = 0; i < fields.length; i++) {
            try {
                fields[i].set(pk, values[fieldNames.indexOf(fields[i].getName())]);
            } catch (IllegalArgumentException e) {
                throw new IdentityTransformException("Unable to convert value for field" + fields[i].getName(), e);
            } catch (IllegalAccessException e) {
                throw new IdentityTransformException("Unable to access field " + fields[i].getName(), e);
            }
        }
        return pk;
    }
}
