/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejb;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.tranql.cache.AlreadyAssociatedException;
import org.tranql.cache.CacheRow;
import org.tranql.cache.CacheRowState;
import org.tranql.cache.CacheTable;
import org.tranql.cache.DuplicateIdentityException;
import org.tranql.cache.InTxCache;
import org.tranql.identity.GlobalIdentity;
import org.tranql.identity.IdentityDefiner;

/**
 * 
 * 
 * @version $Revision: 161 $ $Date: 2005-01-07 01:37:34 -0500 (Fri, 07 Jan 2005) $
 */
public final class ManyToManyCMR implements CMPFieldTransform {
    private final CMPFieldTransform next;
    private final CMPFieldTransform related;
    private final IdentityDefiner relatedIDDefiner;
    private final CacheTable mtmTable;
    private final boolean rightSide;
    
    public ManyToManyCMR(CMPFieldTransform next, CMPFieldTransform related, IdentityDefiner relatedIDDefiner, CacheTable mtmTable, boolean rightSide) {
        this.next = next;
        this.related = related;
        this.relatedIDDefiner = relatedIDDefiner;
        this.mtmTable = mtmTable;
        this.rightSide = rightSide;
    }

    /**
     * Wraps the underlying set so that changes can be detected and notifications sent
     * @param cache the current transaction cache
     * @param row the row being accessed
     * @return the underlying set wrapped for change detection
     */
    public Object get(InTxCache cache, CacheRow row) {
        return new ManyToManyRelationSet(cache, row.getId(), (Set) next.get(cache, row));
    }

    /**
     * Not supported.
     * @throws UnsupportedOperationException
     */
    public void set(InTxCache cache, CacheRow row, Object value) {
        throw new UnsupportedOperationException();
    }

    private class ManyToManyRelationSet extends AbstractRelationSet {
        private final InTxCache cache;
        private final GlobalIdentity id;

        public ManyToManyRelationSet(InTxCache cache, GlobalIdentity id, Set idSet) {
            super(cache, relatedIDDefiner, idSet);
            this.cache = cache;
            this.id = id;
        }

        protected void associate(GlobalIdentity b2) {
            CacheRow b2Row = cache.get(b2);
            if (null == b2Row) {
                b2Row = b2.getTable().emptyRow(b2);
                try {
                    cache.associate(b2Row);
                } catch (AlreadyAssociatedException e) {
                    throw new AssertionError(e);
                }
                relatedIDDefiner.injectIdentity(b2Row);
            }
            Set b2Set = (Set) related.get(cache, b2Row);
            boolean isAdded = b2Set.add(id); 
            if ( isAdded ) {
                GlobalIdentity mtmID = getGlobalIdentity(b2);
                CacheRow row = cache.get(mtmID);
                if ( null != row ) {
                    if ( CacheRowState.NEW == row.getState() ) {
                        return;
                    }
                    // TODO one needs to set the status back to NEW.
                    return;
                }
                row = mtmTable.emptyRow(mtmID);
                if ( rightSide ) {
                    row.set(1, id);
                    row.set(0, b2);
                } else {
                    row.set(0, id);
                    row.set(1, b2);
                }                
                try {
                    mtmTable.addRow(cache, mtmID, row);
                } catch (DuplicateIdentityException e) {
                    throw (IllegalStateException) new IllegalStateException().initCause(e);
                }
            }
        }

        protected void disassociate(GlobalIdentity b2) {
            CacheRow b2Row = cache.get(b2);
            if (null == b2Row) {
                b2Row = b2.getTable().emptyRow(b2);
                try {
                    cache.associate(b2Row);
                } catch (AlreadyAssociatedException e) {
                    throw new AssertionError(e);
                }
                relatedIDDefiner.injectIdentity(b2Row);
            }
            Set b2Set = (Set) related.get(cache, b2Row);
            boolean isRemoved = b2Set.remove(id);
            if ( isRemoved ) {
                GlobalIdentity mtmID = getGlobalIdentity(b2);
                CacheRow row = cache.get(mtmID);
                if ( null == row ) {
                    row = mtmTable.emptyRow(mtmID);
                    if ( rightSide ) {
                        row.set(1, id);
                        row.set(0, b2);
                    } else {
                        row.set(0, id);
                        row.set(1, b2);
                    }                    
                    try {
                        mtmTable.addRow(cache, mtmID, row);
                    } catch (DuplicateIdentityException e) {
                        throw (IllegalStateException) new IllegalStateException().initCause(e);
                    }
                }
                mtmTable.removeRow(cache, mtmID);
            }
        }

        public Iterator iterator() {
            return idSet.iterator();
        }
        
        private GlobalIdentity getGlobalIdentity(GlobalIdentity b2) {
            List list = Arrays.asList(new Object[] {id, b2});
            if ( rightSide ) {
                Collections.reverse(list);  
            }
            return new GlobalIdentity(mtmTable, list);
        }
    }
}
