/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejb;

import java.lang.ref.WeakReference;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.HashMap;
import javax.ejb.EJBException;
import javax.transaction.Synchronization;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.tranql.cache.CacheRow;
import org.tranql.cache.CacheRowState;
import org.tranql.cache.InTxCache;
import org.tranql.identity.GlobalIdentity;
import org.tranql.identity.IdentityTransform;
import org.tranql.identity.IdentityTransformException;

/**
 * An accessor for multi-valued cmr-fields which behaves according to the EJB spec.
 * Specifically, the accessor always returns the same instance for each invocation
 * within a transaction and returns a different instance for different transactions.
 * A snapshot of the collection content is made at the end of the transaction allowing
 * read operations to be performed outside a transaction context.
 *
 * @version $Revision: 156 $ $Date: 2004-12-05 05:00:27 -0500 (Sun, 05 Dec 2004) $
 */
public final class MultiValuedCMRAccessor implements CMPFieldTransform, Synchronization {
    private final CMPFieldTransform next;
    private final TransactionManager txManager;
    private final IdentityTransform idTransform;
    private final Class proxyClass;
    private transient final Map setMapsByTx = Collections.synchronizedMap(new WeakHashMap());

    public MultiValuedCMRAccessor(CMPFieldTransform next, TransactionManager txManager, IdentityTransform idTransform, Class proxyClass) {
        this.next = next;
        this.txManager = txManager;
        this.idTransform = idTransform;
        this.proxyClass = proxyClass;
    }

    public Object get(InTxCache cache, CacheRow row) {
        Transaction tx = getTransaction();
        if (tx == null) {
            throw new IllegalStateException("cmr-getter called when not in a transaction");
        }

        Map setsForTx = getCMRSetsForTx(tx);

        GlobalIdentity id = row.getId();
        CMRWrapper cmrSet = (CMRWrapper) setsForTx.get(id);
        if (cmrSet == null) {
            Set idSet = (Set) next.get(cache, row);
            cmrSet = new CMRWrapper();
            cmrSet.delegate = new CMRSet(idSet, cache);
            setsForTx.put(id, cmrSet);
        }
        return cmrSet;
    }

    public void set(InTxCache cache, CacheRow row, Object value) {
        if (value == null) {
            throw new IllegalArgumentException("null value passed to cmr-setter");
        }

        Transaction tx = getTransaction();
        if (tx == null) {
            throw new IllegalStateException("cmr-setter called when not in a transaction");
        }

        // extract ids from proxies, assuring that they are the right type
        Collection c = (Collection) value;
        ArrayList ids = new ArrayList(c.size());
        for (Iterator i = c.iterator(); i.hasNext();) {
            ids.add(getIdentityAndEnsureNotRemoved(cache, i.next()));
        }

        // update the set with clear/addAll per the spec
        CMRWrapper wrapper = (CMRWrapper) get(cache, row);
        Set idSet = ((CMRSet) wrapper.delegate).idSet;
        idSet.clear();
        idSet.addAll(ids);
    }

    public void beforeCompletion() {
    }

    public void afterCompletion(int status) {
        Transaction tx = getTransaction();
        Map setsInTx = (Map) setMapsByTx.remove(tx);
        for (Iterator i = setsInTx.values().iterator(); i.hasNext();) {
            CMRWrapper cmrWrapper = (CMRWrapper) i.next();
            CMRSet cmrSet = (CMRSet) cmrWrapper.delegate;
            cmrSet.invalidate();
            cmrWrapper.delegate = DEAD_SET;
        }
    }

    private Transaction getTransaction() {
        try {
            return txManager.getTransaction();
        } catch (SystemException e) {
            throw new EJBException("Unable to fetch current transaction", e);
        }
    }

    private GlobalIdentity getIdentity(Object o) {
        if (!proxyClass.isInstance(o)) {
            throw new IllegalArgumentException("Object is not an instance of the cmr type");
        }
        try {
            return idTransform.getGlobalIdentity(o);
        } catch (IdentityTransformException e) {
            throw new EJBException("Unable to extract id from proxy " + o, e);
        }
    }

    private GlobalIdentity getIdentityAndEnsureNotRemoved(InTxCache cache, Object o) {
        GlobalIdentity globalIdentity = getIdentity(o);
        
        // Check that the object to be assigned to this CMR has not been removed.
        CacheRow cacheRow = cache.get(globalIdentity);
        if (null != cacheRow && CacheRowState.REMOVED == cacheRow.getState()) {
            throw new IllegalArgumentException("[EJB2.1 10.3.4.1] Attempt to assign a removed entity object as the value of a cmr-field of another object.");
        }
        
        return globalIdentity;
    }

    private Map getCMRSetsForTx(Transaction tx) {
        Map setsInTx = (Map) setMapsByTx.get(tx);
        if (setsInTx == null) {
            try {
                tx.registerSynchronization(this);
            } catch (Exception e) {
                throw new EJBException("Cannot register for synchronization", e);
            }
            setsInTx = new HashMap();
            setMapsByTx.put(tx, setsInTx);
        }
        return setsInTx;
    }

    private final class CMRSet implements Set {
        private final Set idSet;
        private final List iterators = new ArrayList();
        private final InTxCache cache;

        private CMRSet(Set idSet, InTxCache cache) {
            this.idSet = idSet;
            this.cache = cache;
        }

        public boolean isEmpty() {
            return idSet.isEmpty();
        }

        public int size() {
            return idSet.size();
        }

        public boolean contains(Object o) {
            return idSet.contains(getIdentity(o));
        }

        public boolean containsAll(Collection c) {
            /*
             * Ensure that the collection only contains proxy instances, and
             * then delegate to the underlying identity Set as it may be able
             * to optimize this operation.
             */
            ArrayList ids = new ArrayList(c.size());
            for (Iterator i = c.iterator(); i.hasNext();) {
                ids.add(getIdentity(i.next()));
            }
            return idSet.containsAll(ids);
        }

        public boolean equals(Object obj) {
            /*
             * Follow the contract as defined by Set.equals(Object) - specifically
             * if the parameter is not a Set or does not contain a proxy then
             * return false rather than throwing an IllegalArgumentException.
             * This ensures that if a.equals(b) then b.equals(a) where a is an
             * CMRSet and b is some other implementation of Set.
             */
            if (obj instanceof Set == false) {
                return false;
            }
            Set c = (Set) obj;
            if (idSet.size() != c.size()) {
                return false;
            }
            ArrayList ids = new ArrayList(c.size());
            for (Iterator i = c.iterator(); i.hasNext();) {
                Object o = i.next();
                if (!proxyClass.isInstance(o)) {
                    return false;
                }
                try {
                    ids.add(idTransform.getGlobalIdentity(o));
                } catch (IdentityTransformException e) {
                    throw new EJBException("Unable to extract id from proxy " + o, e);
                }
            }
            return idSet.containsAll(ids);
        }

        /**
         * Implementation of Set.hashCode() which is defined as the sum of the hashCodes of
         * the elements of the set. As the set is perceived to contain proxies, we need to
         * sum those rather than rely on idSet.hashCode() as that would sum the hashCodes of the
         * GlobalIdentity instances it contains. This is probably not very fast, but then I don't
         * expect this to be called often.
         * @return the hashCode for this proxy set
         */
        public int hashCode() {
            int hashCode = 0;
            for (Iterator i = idSet.iterator(); i.hasNext();) {
                GlobalIdentity globalIdentity = (GlobalIdentity) i.next();
                try {
                    Object proxy = idTransform.getDomainIdentity(globalIdentity);
                    hashCode += proxy.hashCode();
                } catch (IdentityTransformException e) {
                    throw new EJBException("Unable to construct proxy for id " + globalIdentity, e);
                }
            }
            return hashCode;
        }

        public Object[] toArray() {
            return toArray(new Object[idSet.size()]);
        }

        public Object[] toArray(Object a[]) {
            if (a.length < idSet.size()) {
                a = (Object[]) Array.newInstance(a.getClass().getComponentType(), idSet.size());
            }
            int idx = 0;
            for (Iterator i = idSet.iterator(); i.hasNext();) {
                GlobalIdentity globalIdentity = (GlobalIdentity) i.next();
                try {
                    a[idx++] = idTransform.getDomainIdentity(globalIdentity);
                } catch (IdentityTransformException e) {
                    throw new EJBException("Unable to construct proxy for id " + globalIdentity, e);
                }
            }
            if (idx < a.length) {
                a[idx] = null;
            }
            return a;
        }

        public Iterator iterator() {
            Iterator i = new CMRIterator(idSet.iterator());
            iterators.add(new WeakReference(i));
            return i;
        }

        public String toString() {
            return idSet.toString();
        }

        public boolean add(Object o) {
            return idSet.add(getIdentityAndEnsureNotRemoved(cache, o));
        }

        public boolean addAll(Collection c) {
            ArrayList ids = new ArrayList(c.size());
            for (Iterator i = c.iterator(); i.hasNext();) {
                ids.add(getIdentityAndEnsureNotRemoved(cache, i.next()));
            }
            return idSet.addAll(ids);
        }

        public void clear() {
            idSet.clear();
        }

        public boolean remove(Object o) {
            return idSet.remove(getIdentity(o));
        }

        public boolean removeAll(Collection c) {
            ArrayList ids = new ArrayList(c.size());
            for (Iterator i = c.iterator(); i.hasNext();) {
                ids.add(getIdentity(i.next()));

            }
            return idSet.removeAll(ids);
        }

        public boolean retainAll(Collection c) {
            ArrayList ids = new ArrayList(c.size());
            for (Iterator i = c.iterator(); i.hasNext();) {
                ids.add(getIdentity(i.next()));

            }
            return idSet.retainAll(ids);
        }

        private void invalidate() {
            for (int i = 0; i < iterators.size(); i++) {
                WeakReference weakReference = (WeakReference) iterators.get(i);
                CMRIterator iterator = (CMRIterator) weakReference.get();
                if (iterator != null) {
                    iterator.i = null;
                }
            }
        }
    }

    private final class CMRIterator implements Iterator {
        private Iterator i;

        public CMRIterator(Iterator i) {
            this.i = i;
        }

        public boolean hasNext() {
            if (i == null) {
                throw new IllegalStateException("Iterator cannot be used outside the transaction context in which it was initialized");
            }
            return i.hasNext();
        }

        public Object next() {
            if (i == null) {
                throw new IllegalStateException("Iterator cannot be used outside the transaction context in which it was initialized");
            }
            GlobalIdentity globalId = (GlobalIdentity) i.next();
            try {
                return idTransform.getDomainIdentity(globalId);
            } catch (IdentityTransformException e) {
                throw new EJBException("Unable to construct proxy for id " + globalId, e);
            }
        }

        public void remove() {
            if (i == null) {
                throw new IllegalStateException("Iterator cannot be used outside the transaction context in which it was initialized");
            }
            i.remove();
        }
    }

    private final class CMRWrapper implements Set {
        private Set delegate;

        private void hosed() {
            try {
                Transaction tx = txManager.getTransaction();
                tx.setRollbackOnly();
            } catch (Exception e) {
                // ignore - the caller will throw the original RuntimeException
                // @todo log this
            }
            delegate = HOSED_SET;
        }

        public boolean add(Object o) {
            try {
                return delegate.add(o);
            } catch (IllegalStateException e) {
                throw e;
            } catch (IllegalArgumentException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public boolean addAll(Collection c) {
            try {
                return delegate.addAll(c);
            } catch (IllegalStateException e) {
                throw e;
            } catch (NullPointerException e) {
                throw e;
            } catch (IllegalArgumentException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public void clear() {
            try {
                delegate.clear();
            } catch (IllegalStateException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public boolean contains(Object o) {
            return delegate.contains(o);
        }

        public boolean containsAll(Collection c) {
            return delegate.containsAll(c);
        }

        public boolean equals(Object o) {
            return delegate.equals(o);
        }

        public int hashCode() {
            return delegate.hashCode();
        }

        public boolean isEmpty() {
            return delegate.isEmpty();
        }

        public Iterator iterator() {
            return delegate.iterator();
        }

        public int size() {
            return delegate.size();
        }

        public boolean remove(Object o) {
            try {
                return delegate.remove(o);
            } catch (IllegalStateException e) {
                throw e;
            } catch (IllegalArgumentException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public boolean removeAll(Collection c) {
            try {
                return delegate.removeAll(c);
            } catch (IllegalStateException e) {
                throw e;
            } catch (NullPointerException e) {
                throw e;
            } catch (IllegalArgumentException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public boolean retainAll(Collection c) {
            try {
                return delegate.retainAll(c);
            } catch (IllegalStateException e) {
                throw e;
            } catch (NullPointerException e) {
                throw e;
            } catch (IllegalArgumentException e) {
                throw e;
            } catch (RuntimeException e) {
                hosed();
                throw e;
            }
        }

        public Object[] toArray() {
            return delegate.toArray();
        }

        public Object[] toArray(Object a[]) {
            return delegate.toArray(a);
        }

        public String toString() {
            return delegate.toString();
        }
    }

    private static final Set DEAD_SET = new Set() {
        private static final String MESSAGE = "[EJB2.1 10.4.2.2] A cmr-field Set cannot be accessed in a transaction context other than that in which it was initially materialized.";

        public boolean add(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean addAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public void clear() {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean contains(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean containsAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean isEmpty() {
            throw new IllegalStateException(MESSAGE);
        }

        public Iterator iterator() {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean remove(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean removeAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean retainAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public int size() {
            throw new IllegalStateException(MESSAGE);
        }

        public Object[] toArray() {
            throw new IllegalStateException(MESSAGE);
        }

        public Object[] toArray(Object a[]) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean equals(Object obj) {
            throw new IllegalStateException(MESSAGE);
        }

        public int hashCode() {
            throw new IllegalStateException(MESSAGE);
        }

        public String toString() {
            throw new IllegalStateException(MESSAGE);
        }
    };

    private static final Set HOSED_SET = new Set() {
        private static final String MESSAGE = "Operations are disabled as a previous Exception has possibly made the underlying relationship inconsistent.";

        public boolean add(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean addAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public void clear() {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean contains(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean containsAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean isEmpty() {
            throw new IllegalStateException(MESSAGE);
        }

        public Iterator iterator() {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean remove(Object o) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean removeAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean retainAll(Collection c) {
            throw new IllegalStateException(MESSAGE);
        }

        public int size() {
            throw new IllegalStateException(MESSAGE);
        }

        public Object[] toArray() {
            throw new IllegalStateException(MESSAGE);
        }

        public Object[] toArray(Object a[]) {
            throw new IllegalStateException(MESSAGE);
        }

        public boolean equals(Object obj) {
            throw new IllegalStateException(MESSAGE);
        }

        public int hashCode() {
            throw new IllegalStateException(MESSAGE);
        }

        public String toString() {
            throw new IllegalStateException(MESSAGE);
        }
    };
    
    private Object readResolve() {
        return new MultiValuedCMRAccessor(next, txManager, idTransform, proxyClass);
    }

}
