/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;

import org.tranql.builder.QueryTileBuilder;
import org.tranql.builder.QueryTileBuilder.QuerySourceFactory;
import org.tranql.ql.Empty;
import org.tranql.ql.Exists;
import org.tranql.ql.From;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.NumericLiteral;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.Select;
import org.tranql.ql.Where;
import org.tranql.ql.queryvisitor.ExecutorVisitor;
import org.tranql.ql.queryvisitor.TraverseAllNodeExecutor;
import org.tranql.schema.AssociationEnd;

/**
 * Transform Empty to exists sub-query clauses.
 * 
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
class EmptyTransformer extends InterRepresentationTransformer {
    private static final EmptySearcher SEARCHER = new EmptySearcher();
    
    public EmptyTransformer(QuerySourceDictionary sourceDictionary) {
        super(sourceDictionary);
    }

    public Object visit(Query node, Object param) throws QueryException {
    	QuerySourceDictionary sourceDictionary = (QuerySourceDictionary) param;
        Collection empties = new ArrayList();
        SEARCHER.visit(node, empties);
        
        Context ctx = new Context(empties, sourceDictionary);
        return super.visit(node, ctx);
    }
    
    public Object visit(Empty node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        // an EMPTY is replaced by an NOT EXISTS sub-query and vice-versa.
        Exists exist = new Exists(!node.isNot());
        
        // build the from clause of the sub-query.
        Path path = node.getPath();
        List thePath = path.getPath();
        Node fromNode = null;
        for (int i = 0; i < thePath.size(); i++) {
            AssociationEnd end = (AssociationEnd) thePath.get(i);
            QuerySource qs = ctx.factory(i + 1, node);
            if (end.isManyToMany()) {
                QuerySource mtmQS = ctx.factoryMTM(node);
                Node joinMTM = new Join(Join.CROSS_JOIN).addChild(mtmQS).addChild(qs);
                fromNode = null == fromNode ? joinMTM : new Join(Join.CROSS_JOIN).addChild(fromNode).addChild(joinMTM);
            } else {
                fromNode = null == fromNode ? qs : new Join(Join.CROSS_JOIN).addChild(fromNode).addChild(qs);
            }
        }

        // build the where clause of the sub-query
        Node whereNode = QueryTileBuilder.buildJoinCondition(new EmptyFactory(ctx, node));

        // assemble the sub-query
        exist.addChild(new Select(false).addChild(new NumericLiteral(BigInteger.ZERO)));
        exist.addChild(new From().addChild(fromNode));
        exist.addChild(new Where().addChild(whereNode));
        
        return exist;
    }
    
    private static class EmptySearcher extends ExecutorVisitor {
        private EmptySearcher() {
            super(new TraverseAllNodeExecutor());
            TraverseAllNodeExecutor allNodeExecutor = (TraverseAllNodeExecutor) executor;
            allNodeExecutor.setQueryVisitor(this);
        }
        
        public Object visit(Empty node, Object param) throws QueryException {
            Collection empties = (Collection) param;
            empties.add(node);
            return null;
        }
    }
    
    private static class EmptyFactory implements QuerySourceFactory {
        private final Context ctx;
        private final Empty empty;
        
        public EmptyFactory(Context ctx, Empty empty) {
            this.ctx = ctx;
            this.empty = empty;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factory(index, empty);
        }
        
        public QuerySource factoryMTM() throws QueryException {
            return ctx.factoryMTM(empty);
        }
        
        public Path getPath() {
            return empty.getPath();
        }
    }
    
    private static class Context {
        private final Collection empties;
        private final IdentityHashMap emptiesQS = new IdentityHashMap();
        
        private Context(Collection empties, QuerySourceDictionary sourceDictionary) throws QueryException {
            this.empties = Collections.unmodifiableCollection(empties);
            
            for (Iterator iter = empties.iterator(); iter.hasNext();) {
                Empty empty = (Empty) iter.next();
                
                Path path = empty.getPath();
                List thePath = path.getPath();
                QuerySource[] sources = new QuerySource[thePath.size() + 2];
                emptiesQS.put(empty, sources);
                
                QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                if (null == qs) {
                    throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                }
                sources[0] = qs;
                
                for (int i = 0; i < thePath.size(); i++) {
                    AssociationEnd end = (AssociationEnd) thePath.get(i);
                    sources[i + 1] = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                    sourceDictionary.registerQuerySource(sources[i + 1]);
                    if (end.isManyToMany()) {
                        sources[i + 2] = new QuerySource(end.getAssociation().getManyToManyEntity(), sourceDictionary.getNextAlias());
                        sourceDictionary.registerQuerySource(sources[i + 2]);
                    }
                }
            }
        }
        
        public Collection getRegisteredEmpties() {
            return empties;
        }
        
        public QuerySource factory(int index, Empty empty) throws QueryException {
            QuerySource[] sources = (QuerySource[]) emptiesQS.get(empty);
            if (null == sources) {
                throw new AssertionError("Unknown Empty=" + empty);
            }
            
            return sources[index];
        }
        
        public QuerySource factoryMTM(Empty empty) throws QueryException {
            QuerySource[] sources = (QuerySource[]) emptiesQS.get(empty);
            if (null == sources) {
                throw new AssertionError("Unknown Empty=" + empty);
            }
            
            return sources[sources.length - 1];
        }

    }
    
}
