/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;

import org.tranql.builder.QueryTileBuilder;
import org.tranql.builder.QueryTileBuilder.QuerySourceFactory;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.Exists;
import org.tranql.ql.From;
import org.tranql.ql.IsNull;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.NumericLiteral;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.RowConstructor;
import org.tranql.ql.Select;
import org.tranql.ql.SubQuery;
import org.tranql.ql.Where;
import org.tranql.ql.queryvisitor.ExecutorVisitor;
import org.tranql.ql.queryvisitor.TraverseAllNodeExecutor;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;

/**
 * 
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
class IsNullTransformer extends InterRepresentationTransformer {
    private static final IsNullSearcher SEARCHER = new IsNullSearcher();
    
    public IsNullTransformer(QuerySourceDictionary sourceDictionary) {
        super(sourceDictionary);
    }
    
    public Object visit(Query node, Object param) throws QueryException {
    	QuerySourceDictionary sourceDictionary = (QuerySourceDictionary) param;
        Collection isNulls = new ArrayList();
        SEARCHER.visit(node, isNulls);
        
        Context ctx = new Context(isNulls, sourceDictionary);
        return super.visit(node, ctx);
    }
    
    public Object visit(IsNull node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        Node content = node.getChild();
        if (false == content instanceof Path) {
            return super.visit(node, param);
        }
        
        Path path = (Path) content;
        List thePath = path.getPath();
        Object lastElement = thePath.get(thePath.size() - 1);
        if (lastElement instanceof Attribute) {
            Attribute attribute = (Attribute) lastElement;

            IsNull isNull = new IsNull(node.isNot());
            
            if (1 == thePath.size()) {
                isNull.addChild(new AttributeReference(path.getRoot(), attribute));
            } else {
                // build the from clause of the sub-query.
                QuerySource lastQuerySource = null;
                Node fromNode = null;
                for (int i = 0; i < thePath.size() - 1; i++) {
                    lastQuerySource = ctx.factory(i + 1, node);
                    fromNode = null == fromNode ? lastQuerySource : new Join(Join.CROSS_JOIN).addChild(fromNode).addChild(lastQuerySource);
                }
                
                // build the where clause of the sub-query
                Node whereNode = QueryTileBuilder.buildJoinCondition(new IsNullFactory(ctx, node));

                // assemble the sub-query
                SubQuery subQuery = new SubQuery();
                subQuery.addChild(new Select(false).addChild(new AttributeReference(lastQuerySource, attribute)));
                subQuery.addChild(new From().addChild(fromNode));
                subQuery.addChild(new Where().addChild(whereNode));
                
                RowConstructor rowConstructor = new RowConstructor();
                rowConstructor.addChild(subQuery);
                
                isNull.addChild(rowConstructor);
            }
            
            return isNull;
        }
        
        // a IS NULL is replaced by an NOT EXISTS sub-query and vice-versa.
        Exists exists = new Exists(!node.isNot());

        // build the from clause of the sub-query.
        Node fromNode = null;
        for (int i = 0; i < thePath.size(); i++) {
            QuerySource qs = ctx.factory(i + 1, node);
            fromNode = null == fromNode ? qs : new Join(Join.CROSS_JOIN).addChild(fromNode).addChild(qs);
        }

        // build the where clause of the sub-query
        Node whereNode = QueryTileBuilder.buildJoinCondition(new IsNullFactory(ctx, node));

        // assemble the sub-query
        exists.addChild(new Select(false).addChild(new NumericLiteral(BigInteger.ZERO)));
        exists.addChild(new From().addChild(fromNode));
        exists.addChild(new Where().addChild(whereNode));

        return exists;
    }
    
    private static class IsNullSearcher extends ExecutorVisitor {
        private IsNullSearcher() {
            super(new TraverseAllNodeExecutor());
            TraverseAllNodeExecutor allNodeExecutor = (TraverseAllNodeExecutor) executor;
            allNodeExecutor.setQueryVisitor(this);
        }
        
        public Object visit(IsNull node, Object param) throws QueryException {
            Collection isNulls = (Collection) param;
            isNulls.add(node);
            return null;
        }
    }
    
    private static class IsNullFactory implements QuerySourceFactory {
        private final Context ctx;
        private final IsNull isNull;
        
        public IsNullFactory(Context ctx, IsNull isNull) {
            this.ctx = ctx;
            this.isNull = isNull;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factory(index, isNull);
        }
        
        public QuerySource factoryMTM() throws QueryException {
            throw new QueryException("Muli valued cmr-path used in a IS NULL clause.");
        }
        
        public Path getPath() {
            return (Path) isNull.getChild();
        }
    }
    
    private static class Context {
        private final Collection isNulls;
        private final IdentityHashMap isNullsQS = new IdentityHashMap();
        
        private Context(Collection isNulls, QuerySourceDictionary sourceDictionary) throws QueryException {
            this.isNulls = isNulls;
            
            for (Iterator iter = isNulls.iterator(); iter.hasNext();) {
                IsNull isNull = (IsNull) iter.next();

                Node content = isNull.getChild();
                if (false == content instanceof Path) {
                    continue;
                }
                
                Path path = (Path) content;
                List thePath = path.getPath();
                QuerySource[] sources = new QuerySource[thePath.size() + 1];
                isNullsQS.put(isNull, sources);
                
                QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                if (null == qs) {
                    throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                }
                sources[0] = qs;
                
                for (int i = 0; i < thePath.size(); i++) {
                    Object o = thePath.get(i);
                    if (false == o instanceof AssociationEnd) {
                        break;
                    }
                    AssociationEnd end = (AssociationEnd) o;
                    sources[i + 1] = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                    sourceDictionary.registerQuerySource(sources[i + 1]);
                }
            }
        }

        public Collection getRegisteredIsNulls() {
            return isNulls;
        }

        public QuerySource factory(int index, IsNull isNull) throws QueryException {
            QuerySource[] sources = (QuerySource[]) isNullsQS.get(isNull);
            if (null == sources) {
                throw new AssertionError("Unknown IsNull=" + isNull);
            }
            
            return sources[index];
        }

    }
    
}
