/**
 *
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.ejbqlcompiler;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.tranql.builder.QueryTileBuilder;
import org.tranql.builder.QueryTileBuilder.QuerySourceFactory;
import org.tranql.ejbqlcompiler.StackNodeRegister.NodeRegister;
import org.tranql.ql.AggregateFunction;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.CountAggregateFunction;
import org.tranql.ql.EntityReference;
import org.tranql.ql.From;
import org.tranql.ql.Join;
import org.tranql.ql.Node;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QuerySourceDictionary;
import org.tranql.ql.Select;
import org.tranql.ql.SubQuery;
import org.tranql.ql.SubQuerySource;
import org.tranql.ql.Where;
import org.tranql.query.QueryCloner;
import org.tranql.schema.Association;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;

/**
 * Transform Path clauses.
 * 
 * @version $Revision: 410 $ $Date: 2006-10-21 20:15:17 -0400 (Sat, 21 Oct 2006) $
 */
class PathTransformer extends AbstractSubQueryAwareTransformer {
    
    public PathTransformer(QuerySourceDictionary sourceDictionary) {
        super(sourceDictionary);
    }

    public Object visit(Query node, Object param) throws QueryException {
    	QuerySourceDictionary sourceDictionary = (QuerySourceDictionary) param;
        StackNodeRegister register = new StackNodeRegister();
        PathSearcher.PATH_SEARCHER.visit(node, register);
        Context ctx = new Context(register.getRegisters(), sourceDictionary);
        
        Query returned = (Query) super.visit(node, ctx);
        
        ctx.push(node);
        if (ctx.isCountMultiPkFields()) {
            SubQuery subQuery = new SubQuery();
            subQuery.addChild(returned.getChild());
            SubQuerySource subQuerySource = new SubQuerySource(subQuery, sourceDictionary.getNextAlias());
            
            Select select = new Select(false);
            AggregateFunction count = new CountAggregateFunction(false);
            Attribute attribute = (Attribute) subQuerySource.getEntity().getAttributes().get(0);
            count.addChild(new Path(subQuerySource, Collections.singletonList(attribute)));
            select.addChild(count);

            Node from = new From().addChild(subQuerySource);
            
            Query to = new Query(returned.getParamTransforms(), returned.getResultTransforms());
            to.addChild(select);
            to.addChild(from);
            ctx.pop();
            return to;
        }
        ctx.pop();
        
        return returned;
    }

    public Object visit(Select node, Object param) throws QueryException {
        Context ctx = (Context) param;

        if (ctx.isCountMultiPkFields()) {
            AggregateFunction function = (AggregateFunction) node.getChild();

            Select select = new Select(true);
            Path path = function.getPath();
            List thePath = path.getPath();
            AssociationEnd end = path.getEndAssociationEnd();
            QuerySource qs = ctx.factoryQS(thePath.size(), path);

            List pkFields = end.getEntity().getPrimaryKeyFields();
            for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
                Attribute att = (Attribute) iter.next();
                select.addChild(new AttributeReference(qs, att));
            }
            return select;
        } else if (false == node.getChild() instanceof Path) {
            return super.visit(node, param);
        }
        
        Select select = new Select(node.isDistinct());
        
        Path path = (Path) node.getChild(); 
        
        List thePath = path.getPath();
        if (path.isEndAttribute()) {
            AliasedEntity aliasedEntity;
            if (1 == thePath.size()) {
                aliasedEntity = path.getRoot();
            } else {
                aliasedEntity = ctx.factoryQS(thePath.size() - 1, path);
            }
            select.addChild(new Path(aliasedEntity, Arrays.asList(new Object[] {path.getEndAttribute()})));
            return select;
        }
        
        AssociationEnd end = path.getEndAssociationEnd();
        QuerySource qs = ctx.factoryQS(thePath.size(), path);
        
        List pkFields = end.getEntity().getPrimaryKeyFields();
        for (Iterator iter = pkFields.iterator(); iter.hasNext();) {
            Attribute att = (Attribute) iter.next();
            select.addChild(new AttributeReference(qs, att));
        }
        
        return select;
    }

    public Object visit(AggregateFunction node, Object param) throws QueryException {
        Context ctx = (Context) param;

        AggregateFunction to = new AggregateFunction(node.getFunctionType(), node.isDistinct());
        return transformPathInAggregateFunction(node.getPath(), ctx, to);        
    }

    private Object transformPathInAggregateFunction(Path path, Context ctx, AggregateFunction to) throws QueryException {
        List pathPath = path.getPath();
        if (path.isEndAttribute()) {
            AliasedEntity aliasedEntity;
            if (1 == pathPath.size()) {
                aliasedEntity = path.getRoot();
            } else {
                aliasedEntity = ctx.factoryQS(pathPath.size() - 1, path);
            }
            to.addChild(new Path(aliasedEntity, Arrays.asList(new Object[] {path.getEndAttribute()})));
            return to;
        }

        AssociationEnd end = path.getEndAssociationEnd();
        QuerySource qs = ctx.factoryQS(pathPath.size(), path);

        Attribute att = (Attribute) end.getEntity().getPrimaryKeyFields().get(0);
        path = new Path(qs, Arrays.asList(new Object[] {att}));

        to.addChild(path);
        return to;
    }
    
    public Object visit(CountAggregateFunction node, Object param) throws QueryException {
        Context ctx = (Context) param;

        CountAggregateFunction to = new CountAggregateFunction(node.isDistinct());
        if (node.isEntityReferenceCounted()) {
            EntityReference entityReference = node.getEntityReference();
            Attribute att = (Attribute) entityReference.getEntity().getEntity().getPrimaryKeyFields().get(0);
            Path path = new Path(entityReference.getEntity(), Arrays.asList(new Object[] {att}));
            to.addChild(path);
            return to;        
        } else {
            return transformPathInAggregateFunction(node.getPath(), ctx, to);
        }
    }
    
    public Object visit(From node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        Set addedSources = new HashSet();
        Node to;
        
        if (null != ctx.getSelectedPath() && ctx.getSelectedPath().isEndAssociationEnd()) {
            Path path = ctx.getSelectedPath();
            List thePath = path.getPath();
            AssociationEnd end = path.getEndAssociationEnd();
            if (end.isManyToMany()) {
                throw new QueryException("Multi valued path selected");
            }

            QuerySource qs = ctx.factoryQS(0, path);
            Node selectedPathFrom = qs;
            addedSources.add(qs.getAlias());
            for (int i = 0; i < thePath.size() - 1; i++) {
                qs = ctx.factoryQS(i + 1, path);
                selectedPathFrom = new Join(Join.CROSS_JOIN).addChild(selectedPathFrom).addChild(qs);
                addedSources.add(qs.getAlias());
            }

            Association.JoinDefinition joinDefinition = end.getAssociation().getJoinDefinition();
            QuerySource lastQS = ctx.factoryQS(thePath.size(), path);
            addedSources.add(lastQS.getAlias());
            Node condition = QueryTileBuilder.buildJoinCondition(joinDefinition, qs, lastQS);
            selectedPathFrom = new Join(Join.LEFT_JOIN).addChild(selectedPathFrom).addChild(lastQS).addChild(condition);

            QuerySourceReplacer sourceReplacer = new QuerySourceReplacer(selectedPathFrom, (QuerySource) path.getRoot());
            to = (Node) node.visit(sourceReplacer, param);
        } else {
            to = (Node) super.visit(node, param);
        }
        
        Collection paths = ctx.getPathsToProcess();
        for (Iterator iter = paths.iterator(); iter.hasNext();) {
            Path path = (Path) iter.next();
            
            List thePath = path.getPath();
            for (int i = 0; i < thePath.size(); i++) {
                Object o = thePath.get(i);
                if (false == o instanceof AssociationEnd) {
                    break;
                }
                
                QuerySource qs = ctx.factoryQS(i + 1, path);
                if (false == addedSources.add(qs.getAlias())) {
                    continue;
                }
                
                to = new Join(Join.CROSS_JOIN).addChild(to).addChild(qs);
            }
        }
    
        return to;
    }
    
    public Object visit(Where node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        Collection paths = ctx.getPathsToProcess();
        Collection factories = new ArrayList();
        for (Iterator iter = paths.iterator(); iter.hasNext();) {
            Path path = (Path) iter.next();
            factories.add(new PathFactory(ctx, path));
        }
        
        Node cond = QueryTileBuilder.buildJoinCondition(factories);

        Path selectedPath = ctx.getSelectedPath();
        if (null != selectedPath && selectedPath.isEndAssociationEnd()) {
            List selectedThePath = selectedPath.getPath();
            for (int i = 0; i < selectedThePath.size() - 1; i++) {
                AssociationEnd end = (AssociationEnd) selectedThePath.get(i);
                Association association = end.getAssociation();
                QuerySource previousQS = ctx.factoryQS(i, selectedPath);
                QuerySource qs = ctx.factoryQS(i + 1, selectedPath);
                Node joinCond = QueryTileBuilder.buildJoinCondition(association.getJoinDefinition(), previousQS, qs);
                cond = null == cond ? joinCond : new BinaryOperation(BinaryOperation.AND).addChild(cond).addChild(joinCond);
            }
        }
        
        if (null == cond) {
            return super.visit(node, param);
        }

        return QueryTileBuilder.buildWhere(node, cond, this, param);
    }

    public Object visit(Path node, Object param) throws QueryException {
        Context ctx = (Context) param;
        
        if (false == ctx.containsPath(node)) {
            return super.visit(node, param);
        }

        AliasedEntity aliasedEntity = node.getRoot();
        List thePath = node.getPath();
        if (1 == thePath.size()) {
            aliasedEntity = node.getRoot();
        } else {
            aliasedEntity = ctx.factoryQS(thePath.size() - 1, node);
        }
        return new Path(aliasedEntity, Arrays.asList(new Object[] {node.getEnd()}));
    }

    private static class QuerySourceReplacer extends QueryCloner {
        private final Node returned;
        private final QuerySource searched;
        private boolean isFound = false;
        private QuerySourceReplacer(Node returned, QuerySource searched) {
            this.returned = returned;
            this.searched = searched;
        }
        public Object visit(QuerySource node, Object param) throws QueryException {
            if (node.getAlias().equals(searched.getAlias())) {
                isFound = true;
                returned.setSibling(node.getSibling());
                return returned;
            } else if (isFound) {
                return null;
            }
            return super.visit(node, param);
        }
    };
    
    private static class PathFactory implements QuerySourceFactory {
        private final Context ctx;
        private final Path path;
        
        public PathFactory(Context ctx, Path path) {
            this.ctx = ctx;
            this.path = path;
        }
        
        public QuerySource factory(int index) throws QueryException {
            return ctx.factoryQS(index, path);
        }
        
        public QuerySource factoryMTM() throws QueryException {
            throw new AssertionError("Muli valued cmr-path has been registered.");
        }
        
        public Path getPath() {
            return path;
        }
    }

    private static class Context extends AbstractSubQueryContext {
        private final IdentityHashMap nodeToSubCtxs = new IdentityHashMap();
        private final Map pathsQS = new HashMap();
        private final QuerySourceDictionary sourceDictionary;
        
        private Context(IdentityHashMap registers, QuerySourceDictionary sourceDictionary) throws QueryException {
            this.sourceDictionary = sourceDictionary;
            
            for (Iterator iter = registers.entrySet().iterator(); iter.hasNext();) {
                Map.Entry entry = (Map.Entry) iter.next();
                Node node = (Node) entry.getKey();
                NodeRegister register = (NodeRegister) entry.getValue();
                push(node);
                processNodeRegister(node, register);
                pop();
            }
        }

        private void processNodeRegister(Node node, NodeRegister register) throws QueryException {
            SubContext subContext = new SubContext();
            subContext.node = node;
            nodeToSubCtxs.put(node, subContext);

            subContext.countMultiPkFields = PathSearcher.PATH_SEARCHER.isFoundCountDistinctMultiPKFields(register);
            
            Path selectedPath = PathSearcher.PATH_SEARCHER.getFoundSelectedPath(register);
            if (null != selectedPath) {
                subContext.selectedPath = selectedPath;
                processPath(sourceDictionary, selectedPath);
            }

            subContext.paths = PathSearcher.PATH_SEARCHER.getFoundPaths(register);
            
            subContext.pathsToProcess = derivePathsToProcess(subContext);

            for (Iterator iter = subContext.pathsToProcess.iterator(); iter.hasNext();) {
                Path path = (Path) iter.next();
                processPath(sourceDictionary, path);
            }
        }

        private Collection derivePathsToProcess(SubContext subContext) throws QueryException {
            Collection filteredPaths;
            if (null != subContext.selectedPath && subContext.selectedPath.isEndAssociationEnd()) {
                filteredPaths = derivePathsToProcessWithSelectedPath(subContext);
            } else if (null != subContext.selectedPath) {
                filteredPaths = new ArrayList(subContext.paths);
                filteredPaths.add(subContext.selectedPath);
            } else {
                filteredPaths = subContext.paths;
            }
            return filteredPaths;
        }

        private Collection derivePathsToProcessWithSelectedPath(SubContext subContext) throws QueryException {
            List selectedPathPath = subContext.selectedPath.getPath(); 
            QuerySource baseSelectedPathQS = factoryQS(selectedPathPath.size(), subContext.selectedPath);
            Collection pathsToProcess = new ArrayList();
            for (int i = 0; i < subContext.paths.size(); i++) {
                Path path = (Path) subContext.paths.get(i);
                if (path.getRoot() == subContext.selectedPath.getRoot()) {
                    List pathPath = path.getPath();
                    if (pathPath.size() - 1 > selectedPathPath.size() && 
                            pathPath.subList(0, selectedPathPath.size()).containsAll(selectedPathPath)) {
                        if (selectedPathPath.size() - 2 > 0) {
                            pathsToProcess.add(new Path(path.getRoot(), pathPath.subList(0, selectedPathPath.size() - 2)));
                        }
                        if (selectedPathPath.size() - 1 < pathPath.size() - 2) {
                            pathsToProcess.add(new Path(baseSelectedPathQS, pathPath.subList(selectedPathPath.size() - 1, pathPath.size() - 1)));
                        }
                    } else {
                        pathsToProcess.add(path);
                    }
                } else {
                    pathsToProcess.add(path);
                }
            }
            return pathsToProcess;
        }

        public Collection getPathsToProcess() {
            return getSubContext().pathsToProcess;
        }

        public Path getSelectedPath() {
            return getSubContext().selectedPath;
        }
        
        public boolean containsPath(Path path) {
            return getSubContext().containsPath(path);
        }
        
        public QuerySource factoryQS(int index, Path path) throws QueryException {
            return getSubContext().factoryQS(index, path);
        }
        
        public void setCountMultiPkFields(boolean countMultiPkFields) {
            getSubContext().countMultiPkFields = countMultiPkFields;
        }

        public boolean isCountMultiPkFields() {
            return getSubContext().countMultiPkFields;
        }

        private void processPath(QuerySourceDictionary sourceDictionary, Path path) throws QueryException {
            getSubContext().processPath(sourceDictionary, path);
        }

        private SubContext getSubContext() {
            return (SubContext) nodeToSubCtxs.get(peek());
        }

        private class SubContext {
            private Node node;
            private boolean countMultiPkFields;
            private Path selectedPath;
            private List paths;
            private Collection pathsToProcess;

            public boolean containsPath(Path path) {
                return null != paths ? paths.contains(path) : false;
            }
            
            public QuerySource factoryQS(int index, Path path) throws QueryException {
                List key = newKey(path);
                key.addAll(path.getPath().subList(0, index));
                
                QuerySource qs = (QuerySource) pathsQS.get(key);
                if (null == qs) {
                    throw new AssertionError("Unknown Path=" + path);
                }
                
                return qs;
            }

            private void processPath(QuerySourceDictionary sourceDictionary, Path path) throws QueryException {
                List key = newKey(path);
                
                QuerySource qs = sourceDictionary.retrieveQuerySource(path.getRoot().getAlias());
                if (null == qs) {
                    throw new QueryException("Unknown alias=" + path.getRoot().getAlias());
                }
                if (false == pathsQS.containsKey(key)) {
                    pathsQS.put(new ArrayList(key), qs);
                }
                
                List pathPath = path.getPath();
                for (Iterator iter = pathPath.iterator(); iter.hasNext();) {
                    Object o = iter.next();
                    if (o instanceof AssociationEnd) {
                        AssociationEnd end = (AssociationEnd) o;
                        key.add(end);
                        qs = (QuerySource) pathsQS.get(key);
                        if (null == qs) {
                            qs = new QuerySource(end.getEntity(), sourceDictionary.getNextAlias());
                            sourceDictionary.registerQuerySource(qs);
                            pathsQS.put(new ArrayList(key), qs);
                        } 
                    }
                }
            }
            
            private List newKey(Path path) {
                List key = new ArrayList();
                key.add(node);
                key.add(path.getRoot().getAlias());
                return key;
            }
        }
    }

}
