/**
 * 
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.intertxcache;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.transaction.Status;
import javax.transaction.Transaction;

import org.tranql.cache.CacheRow;
import org.tranql.identity.GlobalIdentity;

/**
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class BasicCache implements Cache {
    private final Map cache;
    
    public BasicCache() {
        this.cache = new HashMap();
    }
    
    public void transactionStart(Transaction tx) throws CacheException {
    }
    
    public void transactionEnd(Transaction tx, int status, Collection rows) throws CacheException {
        if (status == Status.STATUS_ROLLEDBACK) {
            synchronized (cache) {
                for (Iterator iter = rows.iterator(); iter.hasNext();) {
                    CacheRowEntry row = (CacheRowEntry) iter.next();
                    remove(tx, row.getId());
                }
            }
        } else if (status == Status.STATUS_COMMITTED) {
            synchronized (cache) {
                for (Iterator iter = rows.iterator(); iter.hasNext();) {
                    CacheRowEntry row = (CacheRowEntry) iter.next();
                    if (row.isRemoved()) {
                        remove(tx, row.getId());
                    } else {
                        CacheRowEntry entry = (CacheRowEntry) cache.get(row.getId());
                        if (null != entry) {
                            entry.merge(row);
                        } else {
                            cache.put(row.getId(), row);
                        }
                    }
                }
            }
        } else {
            throw new IllegalArgumentException("Invalid status " + status);
        }
    }

    public void put(Transaction tx, CacheRow row) throws CacheException {
        synchronized (cache) {
            CacheRowEntry entry = (CacheRowEntry) cache.get(row.getId());
            if (null != entry) {
                entry.merge(row);
            } else {
                cache.put(row.getId(), new CacheRowEntry(row));
            }
        }
    }

    public void set(Transaction tx, CacheRow row, int[] indices) throws CacheException {
        CacheRowEntry entry = getEntry(row);
        if (null == entry) {
            return;
        }
        entry.set(row, indices);
    }

    public CacheRow get(Transaction tx, GlobalIdentity id) throws CacheException {
        CacheRowEntry entry = getEntry(id);
        if (null == entry) {
            return null;
        }
        return entry.buildCacheRow();
    }
    
    public boolean get(Transaction tx, CacheRow target, int[] indices) throws CacheException {
        CacheRowEntry entry = getEntry(target);
        if (null == entry) {
            return false;
        }
        return entry.get(target, indices);
    }

    public void remove(Transaction tx, GlobalIdentity id) throws CacheException {
        release(id);
    }
    
    public void release(GlobalIdentity id) throws CacheException {
        synchronized (cache) {
            cache.remove(id);
        }
    }
    
    public void clean() {
        synchronized (cache) {
            cache.clear();
        }
    }
    
    private CacheRowEntry getEntry(CacheRow row) {
        return getEntry(row.getId());
    }
    
    private CacheRowEntry getEntry(GlobalIdentity id) {
        synchronized (cache) {
            return (CacheRowEntry) cache.get(id);
        }
    } 
}