/**
 * 
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.intertxcache;

import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import org.tranql.cache.CacheRow;
import org.tranql.cache.CacheRowState;
import org.tranql.cache.CacheTable;
import org.tranql.identity.GlobalIdentity;

/**
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public final class FrontEndCacheDelegate implements FrontEndCache, Serializable {
    private transient Map entityNameToDelegate;

    public FrontEndCacheDelegate() {
        entityNameToDelegate = new HashMap();
    }
    
    public void addFrontEndCache(String entityName, FrontEndCache delegate) {
        entityNameToDelegate.put(entityName, delegate);
    }

    public void removeFrontEndCache(String entityName) {
        entityNameToDelegate.remove(entityName);
    }

    public boolean get(CacheRow target, int[] indices) throws CacheException {
        if (CacheRowState.UNDER_CONSTRUCTION == target.getState()) {
            return false;
        }

        boolean allLoaded = true;
        for (int i = 0; i < indices.length && allLoaded; i++) {
            int index = indices[i];
            if (false == target.isLoaded(index)) {
                allLoaded = false;
            }
        }
        if (allLoaded) {
            return true;
        }
        
        FrontEndCache delegate = getTarget(target.getId());
        return delegate.get(target, indices);
    }

    public CacheRow get(GlobalIdentity id) throws CacheException {
        FrontEndCache delegate = getTarget(id);
        return delegate.get(id);
    }

    public void put(CacheRow row) throws CacheException {
        if (CacheRowState.UNDER_CONSTRUCTION == row.getState()) {
            return;
        }

        FrontEndCache delegate = getTarget(row.getId());
        // This is a MTM row.
        if (null == delegate) {
            return;
        }
        delegate.put(row);
    }

    public void remove(GlobalIdentity id) throws CacheException {
        FrontEndCache delegate = getTarget(id);
        // This is a MTM row.
        if (null == delegate) {
            return;
        }
        delegate.remove(id);
    }

    public void set(CacheRow row, int[] indices) throws CacheException {
        if (CacheRowState.UNDER_CONSTRUCTION == row.getState()) {
            return;
        }

        FrontEndCache delegate = getTarget(row.getId());
        delegate.set(row, indices);
    }

    private FrontEndCache getTarget(GlobalIdentity id) {
        return getTarget(id.getTable());
    }

    private FrontEndCache getTarget(CacheTable cacheTable) {
        return (FrontEndCache) entityNameToDelegate.get(cacheTable.getName());
    }
    
    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        entityNameToDelegate = new HashMap();
    }
}