/**
 * 
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.intertxcache;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.transaction.Transaction;

import org.tranql.cache.CacheRow;
import org.tranql.identity.GlobalIdentity;

/**
 *
 * @version $Revision: 342 $ $Date: 2006-06-03 18:17:22 -0400 (Sat, 03 Jun 2006) $
 */
public class LRUCache implements Cache {
    private final Cache next;
    private final Map cache;
    private final int maxSize;
    private final Node head;
    private final Node tail;
    
    public LRUCache(int maxSize, Cache next) {
        this.next = next;
        this.maxSize = maxSize;

        cache = new HashMap();
        head = new Node(null);
        tail = new Node(null);
        head.setDown(tail);
    }
    
    public void transactionStart(Transaction tx) throws CacheException {
        next.transactionStart(tx);
    }
    
    public void transactionEnd(Transaction tx, int status, Collection rows) throws CacheException {
        synchronized (cache) {
            for (Iterator iter = rows.iterator(); iter.hasNext();) {
                CacheRowEntry row = (CacheRowEntry) iter.next();
                Node node = (Node) cache.get(row.getId());
                if (null == node) {
                    node = new Node(row.getId());
                    cache.put(row.getId(), node);
                    if (maxSize < cache.size()) {
                        Node last = tail.up;
                        cache.remove(last.id);
                        next.release(last.id);
                        last.remove();
                    }
                }
                setHead(node);
            }
        }
        next.transactionEnd(tx, status, rows);
    }

    public void put(Transaction tx, CacheRow row) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.get(row.getId());
            if (null == node) {
                node = new Node(row.getId());
                cache.put(row.getId(), node);
                if (maxSize < cache.size()) {
                    Node last = tail.up;
                    cache.remove(last.id);
                    next.release(last.id);
                    last.remove();
                }
            }
        }
        setHead(node);
        next.put(tx, row);
    }

    public void set(Transaction tx, CacheRow row, int[] indices) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.get(row.getId());
        }
        if (null == node) {
            return;
        }
        setHead(node);
        next.set(tx, row, indices);
    }

    public CacheRow get(Transaction tx, GlobalIdentity id) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.get(id);
        }
        if (null == node) {
            return null;
        }
        setHead(node);
        return next.get(tx, id);
    }
    
    public boolean get(Transaction tx, CacheRow target, int[] indices) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.get(target.getId());
        }
        if (null == node) {
            return false;
        }
        setHead(node);
        return next.get(tx, target, indices);
    }

    public void remove(Transaction tx, GlobalIdentity id) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.remove(id);
        }
        if (null == node) {
            return;
        }
        node.remove();
        next.remove(tx, id);
    }
    
    public void release(GlobalIdentity id) throws CacheException {
        Node node;
        synchronized (cache) {
            node = (Node) cache.remove(id);
        }
        if (null == node) {
            return;
        }
        node.remove();
        next.release(id);
    }
    
    public void clean() throws CacheException {
        synchronized (cache) {
            cache.clear();
            if (tail != head.down) {
                head.down.remove();
                tail.up.remove();
            }
            head.setDown(tail);
        }
        next.clean();
    }
    
    private void setHead(Node node) {
        synchronized (head) {
            node.remove();
            node.setDown(head.down);
            head.setDown(node);
        }
    }
    
    private static class Node {
        public final GlobalIdentity id;
        public Node up;
        public Node down;

        private Node(GlobalIdentity id) {
            this.id = id;
        }
        
        public void setDown(Node down) {
            this.down = down;
            if (null != down) {
                down.up = this;
            }
        }
        
        public void remove() {
            if (null != up) {
                up.setDown(down);
                up = null;
            }
            down = null;
        }
    }
}