/**
 * 
 *  Copyright 2005 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.intertxcache;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.transaction.Status;
import javax.transaction.Transaction;

import org.tranql.cache.CacheRow;
import org.tranql.identity.GlobalIdentity;

/**
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class MultiVersionCache implements Cache {
    private static final Long MAX_VERSION = new Long(Long.MAX_VALUE); 
    
    private final Object nextVersionLock;
    private long nextVersion;
    private final Map cache;
    private final Map previousVersions;
    private final Object infoLock;
    private final Map txToTxInfo;
    private final TreeMap versionToTxInfo;
    
    public MultiVersionCache() {
        nextVersionLock = new Object();
        cache = new HashMap();
        previousVersions = new HashMap();
        infoLock = new Object();
        txToTxInfo = new HashMap();
        versionToTxInfo = new TreeMap();
    }
    
    public void transactionStart(Transaction tx) throws CacheException {
    }

    public void transactionEnd(Transaction tx, int status, Collection rows) throws CacheException {
        Long version = getTxVersion(tx);
        if (status == Status.STATUS_COMMITTED) {
            synchronized (cache) {
                for (Iterator iter = rows.iterator(); iter.hasNext();) {
                    CacheRowEntry row = (CacheRowEntry) iter.next();
                    VersionedEntry entry;
                    if (row.isRemoved()) {
                        entry = (VersionedEntry) cache.remove(row.getId());
                    } else {
                        entry = (VersionedEntry) cache.get(row.getId());
                        if (null != entry) {
                            entry = new VersionedEntry(entry, row, version);
                        } else {
                            entry = new VersionedEntry(row, version);
                        }
                        entry = (VersionedEntry) cache.put(row.getId(), entry);
                    }
                    if (null != entry) {
                        addToPreviousVersions(row.getId(), entry);
                    }
                }
            }
        }
        TxInfo info;
        synchronized (infoLock) {
            info = (TxInfo) txToTxInfo.remove(tx);
            versionToTxInfo.remove(info.version);
        }
        for (Iterator iter = info.entries.iterator(); iter.hasNext();) {
            VersionedEntry entry = (VersionedEntry) iter.next();
            entry.removeTransaction(tx);
        }
    }

    public void put(Transaction tx, CacheRow row) throws CacheException {
        synchronized (cache) {
            Long version = getNextVersion();
            VersionedEntry entry = (VersionedEntry) cache.get(row.getId());
            if (null != entry) {
                entry = new VersionedEntry(entry, row, version);
            } else {
                entry = new VersionedEntry(row, version);
            }
            entry = (VersionedEntry) cache.put(row.getId(), entry);
            if (null != entry) {
                addToPreviousVersions(row.getId(), entry);
            }
        }
    }

    public void set(Transaction tx, CacheRow row, int[] indices) throws CacheException {
        VersionedEntry entry =  getEntry(tx, row.getId());
        if (null == entry) {
            return;
        }
        entry.set(row, indices);
    }

    public CacheRow get(Transaction tx, GlobalIdentity id) throws CacheException {
        VersionedEntry entry =  getEntry(tx, id);
        if (null == entry) {
            return null;
        }
        return entry.buildCacheRow();
    }

    public boolean get(Transaction tx, CacheRow target, int[] indices) throws CacheException {
        VersionedEntry entry =  getEntry(tx, target.getId());
        if (null == entry) {
            return false;
        }
        return entry.get(target, indices);
    }

    public void remove(Transaction tx, GlobalIdentity id) throws CacheException {
        synchronized (cache) {
            cache.remove(id);
        }
    }

    public void release(GlobalIdentity id) throws CacheException {
        synchronized (cache) {
            cache.remove(id);
        }
    }

    public void clean() throws CacheException {
        synchronized (cache) {
            cache.clear();
        }
    }

    private void addToPreviousVersions(GlobalIdentity id, VersionedEntry entry) {
        Map associatedTx;
        synchronized (infoLock) {
            associatedTx = versionToTxInfo.tailMap(new Long(entry.version.longValue() + 1));
        }
        if (associatedTx.isEmpty()) {
            return;
        }
        Collection txs = new ArrayList(associatedTx.values());
        entry.nbTx = txs.size();
        SortedMap entries;
        synchronized (previousVersions) {
            entries = (SortedMap) previousVersions.get(id);
            if (null == entries) {
                entries = new TreeMap();
                previousVersions.put(id, entries);
            }
        }
        synchronized (entries) {
            entries.put(entry.version, entry);
        }
        
        for (Iterator iter = txs.iterator(); iter.hasNext();) {
            TxInfo info = (TxInfo) iter.next();
            info.entries.add(entry);
        }
    }

    private Long getTxVersion(Transaction tx) throws CacheException {
        synchronized (infoLock) {
            TxInfo info = (TxInfo) txToTxInfo.get(tx);
            if (null == info) {
                Long version = getNextVersion();
                info = new TxInfo(tx, version);
                txToTxInfo.put(tx, info);
                versionToTxInfo.put(version, info);
            }
            return info.version;
        }
    }

    private Long getNextVersion() {
        synchronized (nextVersionLock) {
            return new Long(nextVersion++);
        }
    }
    
    private VersionedEntry getEntry(Transaction tx, GlobalIdentity id) throws CacheException {
        Long version;
        if (null != tx) {
            version = getTxVersion(tx);
        } else {
            version = MAX_VERSION;
        }
        VersionedEntry entry;
        synchronized (cache) {
            entry = (VersionedEntry) cache.get(id);
        }
        if (null == entry || version.longValue() < entry.version.longValue()) {
            SortedMap entries;
            synchronized (previousVersions) {
                entries = (SortedMap) previousVersions.get(id);
            }
            if (null == entries) {
                return null;
            }
            synchronized (entries) {
                entries = entries.headMap(new Long(version.longValue() + 1));
                if (entries.isEmpty()) {
                    return null;
                }
                entry = (VersionedEntry) entries.get(entries.lastKey());
            }
        }
        return entry;
    }
    
    private void removeTx(Transaction tx) {
        TxInfo info;
        synchronized (infoLock) {
            info = (TxInfo) txToTxInfo.remove(tx);
            versionToTxInfo.remove(info.version);
        }
        for (Iterator iter = info.entries.iterator(); iter.hasNext();) {
            VersionedEntry entry = (VersionedEntry) iter.next();
            entry.removeTransaction(tx);
        }
    }

    private static class TxInfo {
        public final Set entries;
        public final Transaction tx;
        public final Long version;
        
        public TxInfo(Transaction tx, Long version) {
            this.tx = tx;
            this.version = version;
            entries = new HashSet();
        }
    }
    
    private class VersionedEntry extends CacheRowEntry implements Comparable {
        public final Long version;
        public int nbTx;

        public VersionedEntry(CacheRow row, Long version) {
            super(row);
            this.version = version;
        }

        public VersionedEntry(CacheRowEntry row, Long version) {
            super(row.buildCacheRow());
            this.version = version;
        }

        public VersionedEntry(VersionedEntry prototype, CacheRowEntry row, Long version) {
            super(prototype.buildCacheRow());
            this.version = version;
            merge(row);
        }

        public VersionedEntry(VersionedEntry prototype, CacheRow row, Long version) {
            super(prototype.buildCacheRow());
            this.version = version;
            merge(row);
        }

        public int compareTo(Object o) {
            VersionedEntry entry = (VersionedEntry) o;
            long delta = version.longValue() - entry.version.longValue();
            if (0 == delta) {
                return 0;
            } else if (delta < 0) {
                return -1;
            } else {
                return 1;
            }
        }
        
        public void removeTransaction(Transaction tx) {
            nbTx--;
            if (nbTx < 1) {
                synchronized (previousVersions) {
                    previousVersions.remove(id);
                }
            }
        }
    }
}