/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.jdbc;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import javax.sql.DataSource;

import org.apache.commons.logging.LogFactory;
import org.tranql.cache.InTxCache;
import org.tranql.field.FieldTransformException;
import org.tranql.field.Row;
import org.tranql.ql.Query;
import org.tranql.ql.QueryException;
import org.tranql.query.QueryCommand;
import org.tranql.query.ResultHandler;

/**
 *
 *
 * @version $Revision: 432 $ $Date: 2006-12-12 00:17:36 -0500 (Tue, 12 Dec 2006) $
 */
public class JDBCQueryCommand implements QueryCommand {
	private final static org.apache.commons.logging.Log log = LogFactory.getLog(JDBCQueryCommand.class);

    private final Query query;
    private final DataSource ds;
    private final String sqlText;
    private final InputBinding[] paramBindings;
    private final ResultBinding[] resultBindings;

    public JDBCQueryCommand(Query query, DataSource ds, String sqlText, InputBinding[] paramBindings, ResultBinding[] resultBindings) {
        this.query = query;
        this.ds = ds;
        this.sqlText = sqlText;
        this.paramBindings = paramBindings;
        this.resultBindings = resultBindings;
    }

    public Query getQuery() {
        return query;
    }

    public Object execute(InTxCache cache, ResultHandler handler, Row params, Object arg) throws QueryException {
        Connection c = null;
        PreparedStatement ps = null;
        ResultSet rs = null;
        try {
            c = ds.getConnection();
            ps = c.prepareStatement(sqlText);
            
            if (log.isTraceEnabled())
                log.trace("Prepared SQL Statement=\""+sqlText+"\"");
            
            for (int i = 0; i < paramBindings.length; i++) {
                paramBindings[i].setValue(ps, params);
                if (log.isTraceEnabled())
                    log.trace("  Setting parm["+i+"] "+params.get(i).toString());
            }

            rs = ps.executeQuery();
            
            Row result = new Row(new Object[resultBindings.length]);
            int rows = 0;
            while (rs.next()) {
                // @todo how do we know how big this should be?
                for (int i=0; i < resultBindings.length; i++) {
                    resultBindings[i].getValue(rs, result);
                }
                arg = handler.fetched(result, arg);
                rows++;
            }
            
            if (log.isTraceEnabled())
            	log.trace("  "+rows+" were returned.");
            
            arg = handler.endFetched(arg);
            return arg;
        } catch (SQLException e) {
            throw new QueryException("Error executing statement: " + sqlText, e);
        } catch (FieldTransformException e) {
            throw new QueryException(e);
        } finally {
            JDBCUtil.close(c, ps, rs);
        }
    }

    public String getSQLText() {
        return sqlText;
    }

    public InputBinding[] getParamBindings() {
        return (InputBinding[])paramBindings.clone();
    }

    public ResultBinding[] getResultBindings() {
        return (ResultBinding[])resultBindings.clone();
    }
    
    public String toString() {
        return "JDBCQueryCommand sql=[" + sqlText + "]";
    }
}
