/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.jdbc.binding;

import java.lang.reflect.Constructor;
import java.math.BigDecimal;
import java.sql.Blob;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryException;
import org.tranql.sql.Column;
import org.tranql.sql.FKColumn;
import org.tranql.sql.TypeConverter;
import org.tranql.sql.jdbc.InputBinding;
import org.tranql.sql.jdbc.OutputBinding;
import org.tranql.sql.jdbc.ResultBinding;
import org.tranql.sql.typeconverter.IdentityTypeConverter;
import org.tranql.sql.typeconverter.SerializableConverter;

/**
 *
 *
 * @version $Revision: 323 $ $Date: 2006-02-19 01:29:24 -0500 (Sun, 19 Feb 2006) $
 */
public class BindingFactory {
    private static final Map CTR_BY_JAVA_CLASS = new HashMap();

    static {
        Class[] ctrArgs = {Integer.TYPE, QueryBinding.class};
        try {
            CTR_BY_JAVA_CLASS.put(Boolean.TYPE, BooleanBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Boolean.class, BooleanBinding.class.getConstructor(ctrArgs));

            CTR_BY_JAVA_CLASS.put(Byte.TYPE, ByteBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Byte.class, ByteBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Short.TYPE, ShortBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Short.class, ShortBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Integer.TYPE, IntegerBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Integer.class, IntegerBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Long.TYPE, LongBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Long.class, LongBinding.class.getConstructor(ctrArgs));

            CTR_BY_JAVA_CLASS.put(Float.TYPE, FloatBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Float.class, FloatBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Double.TYPE, DoubleBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Double.class, DoubleBinding.class.getConstructor(ctrArgs));

            CTR_BY_JAVA_CLASS.put(char.class, CharBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Character.TYPE, CharBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(String.class, StringBinding.class.getConstructor(ctrArgs));
            
            CTR_BY_JAVA_CLASS.put(Date.class, DateBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(java.sql.Date.class, SQLDateBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Time.class, TimeBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(Timestamp.class, TimestampBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(BigDecimal.class, BigDecimalBinding.class.getConstructor(ctrArgs));

            CTR_BY_JAVA_CLASS.put(Blob.class, BlobBinding.class.getConstructor(ctrArgs));
            CTR_BY_JAVA_CLASS.put(byte[].class, ByteArrayBinding.class.getConstructor(ctrArgs));
        } catch (Exception e) {
            throw new ExceptionInInitializerError(e);
        }
    }

    private static final Map CTR_BY_SQL_TYPE = new HashMap();
    
    static {
        Class[] ctrArgs = {Integer.TYPE, QueryBinding.class};
        try {
            CTR_BY_SQL_TYPE.put(new Integer(Types.BOOLEAN), BooleanBinding.class.getConstructor(ctrArgs));

            CTR_BY_SQL_TYPE.put(new Integer(Types.TINYINT), ByteBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.SMALLINT), ShortBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.INTEGER), IntegerBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.BIGINT), LongBinding.class.getConstructor(ctrArgs));

            CTR_BY_SQL_TYPE.put(new Integer(Types.FLOAT), FloatBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.DOUBLE), DoubleBinding.class.getConstructor(ctrArgs));

            CTR_BY_SQL_TYPE.put(new Integer(Types.CHAR), CharBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.VARCHAR), StringBinding.class.getConstructor(ctrArgs));

            CTR_BY_SQL_TYPE.put(new Integer(Types.DATE), DateBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.TIME), TimeBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.TIMESTAMP), TimestampBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.DECIMAL), BigDecimalBinding.class.getConstructor(ctrArgs));

            CTR_BY_SQL_TYPE.put(new Integer(Types.BLOB), BlobBinding.class.getConstructor(ctrArgs));
            CTR_BY_SQL_TYPE.put(new Integer(Types.LONGVARBINARY), ByteArrayBinding.class.getConstructor(ctrArgs));
        } catch (Exception e) {
            throw new ExceptionInInitializerError(e);
        }
    }

    public static InputBinding getInputBinding(int index, QueryBinding transform) throws QueryException {
        BindingInfo bindingInfo = getBindingInfo(transform);

        InputBinding inputBinding;
        try {
            inputBinding = (InputBinding) bindingInfo.ctr.newInstance(new Object[]{new Integer(index), transform});
        } catch (Exception e) {
            // should not happen
            throw new QueryException("Unable to create output binding for class " + transform.getFieldClass(), e);
        }
        return new TypeConverterBinding(transform, inputBinding, bindingInfo.typeConverter);
    }

    public static OutputBinding getOutputBinding(int index, QueryBinding transform) throws QueryException {
        BindingInfo bindingInfo = getBindingInfo(transform);

        OutputBinding outputBinding;
        try {
            outputBinding = (OutputBinding) bindingInfo.ctr.newInstance(new Object[]{new Integer(index), transform});
        } catch (Exception e) {
            // should not happen
            throw new QueryException("Unable to create output binding for class " + transform.getFieldClass(), e);
        }
        return new TypeConverterBinding(transform, outputBinding, bindingInfo.typeConverter);
    }

    public static ResultBinding getResultBinding(int index, QueryBinding transform) throws QueryException {
        BindingInfo bindingInfo = getBindingInfo(transform);

        ResultBinding resultBinding;
        try {
            resultBinding = (ResultBinding) bindingInfo.ctr.newInstance(new Object[]{new Integer(index), transform});
        } catch (Exception e) {
            // should not happen
            throw new QueryException("Unable to create result binding for class " + transform.getFieldClass(), e);
        }
        return new TypeConverterBinding(transform, resultBinding, bindingInfo.typeConverter);
    }

    public static BindingInfo getBindingInfo(QueryBinding transform) {
        TypeConverter typeConverter = IdentityTypeConverter.CONVERTER;

        Constructor ctr;
        if (transform.isUnknownBinding()) {
            ctr = (Constructor) CTR_BY_JAVA_CLASS.get(transform.getFieldClass());
            if (null == ctr) {
                typeConverter = SerializableConverter.CONVERTER;
                ctr = (Constructor) CTR_BY_JAVA_CLASS.get(Blob.class);
            }
        } else {
            boolean isSQLTypeSet;
            int sqlType = -1;
            if (transform.isAttributeBinding()) {
                Column column = (Column) transform.getAttribute();
                isSQLTypeSet = column.isSQLTypeSet();
                if (column.isSQLTypeSet()) {
                    sqlType = column.getSQLType();
                }
                if (column.isTypeConverterSet()) {
                    typeConverter = column.getTypeConverter();
                } else {
                    typeConverter = IdentityTypeConverter.CONVERTER;
                }
            } else {
                FKColumn column = (FKColumn) transform.getFKAttribute();
                isSQLTypeSet = column.isSQLTypeSet();
                if (column.isSQLTypeSet()) {
                    sqlType = column.getSQLType();
                }
                if (column.isTypeConverterSet()) {
                    typeConverter = column.getTypeConverter();
                } else {
                    typeConverter = IdentityTypeConverter.CONVERTER;
                }
            }
            if (isSQLTypeSet) {
                ctr = (Constructor) CTR_BY_SQL_TYPE.get(new Integer(sqlType));
                if (null == ctr) {
                    throw new IllegalArgumentException("SQL type " + sqlType + " is not defined.");
                }
            } else {
                ctr = (Constructor) CTR_BY_JAVA_CLASS.get(transform.getFieldClass());
                if (null == ctr) {
                    typeConverter = SerializableConverter.CONVERTER;
                    ctr = (Constructor) CTR_BY_JAVA_CLASS.get(Blob.class);
                }
            }
        }
        return new BindingInfo(ctr, typeConverter);
    }

    public static class BindingInfo {
        private Constructor ctr;
        private TypeConverter typeConverter;
        BindingInfo(Constructor ctr, TypeConverter typeConverter) {
            this.ctr = ctr;
            this.typeConverter = typeConverter;
        }
        public boolean isLob() {
            return CTR_BY_SQL_TYPE.get(new Integer(Types.BLOB)) == ctr;
        }
    }
}
