/**
 *
 *  Copyright 2004 Jeremy Boynes
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.tranql.sql.sql92;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.tranql.ql.AggregateFunction;
import org.tranql.ql.AliasedEntity;
import org.tranql.ql.Assignment;
import org.tranql.ql.AttributeReference;
import org.tranql.ql.Between;
import org.tranql.ql.BinaryOperation;
import org.tranql.ql.BooleanLiteral;
import org.tranql.ql.ConditionalAssignment;
import org.tranql.ql.CountAggregateFunction;
import org.tranql.ql.Delete;
import org.tranql.ql.DerivedTable;
import org.tranql.ql.Empty;
import org.tranql.ql.EntityReference;
import org.tranql.ql.Exists;
import org.tranql.ql.FKAttributeReference;
import org.tranql.ql.From;
import org.tranql.ql.Function;
import org.tranql.ql.In;
import org.tranql.ql.Insert;
import org.tranql.ql.IsNull;
import org.tranql.ql.Join;
import org.tranql.ql.Like;
import org.tranql.ql.MalformedQueryException;
import org.tranql.ql.MemberOf;
import org.tranql.ql.NoOpQuery;
import org.tranql.ql.Node;
import org.tranql.ql.Not;
import org.tranql.ql.NumericLiteral;
import org.tranql.ql.OrderBy;
import org.tranql.ql.ParameterReference;
import org.tranql.ql.Path;
import org.tranql.ql.Query;
import org.tranql.ql.QueryBinding;
import org.tranql.ql.QueryException;
import org.tranql.ql.QuerySource;
import org.tranql.ql.QueryVisitor;
import org.tranql.ql.RowConstructor;
import org.tranql.ql.Select;
import org.tranql.ql.SetList;
import org.tranql.ql.StringLiteral;
import org.tranql.ql.SubQuery;
import org.tranql.ql.SubQuerySource;
import org.tranql.ql.TableConstructor;
import org.tranql.ql.UnaryOperator;
import org.tranql.ql.Update;
import org.tranql.ql.Where;
import org.tranql.query.CommandFactory;
import org.tranql.query.QueryCommand;
import org.tranql.query.UpdateCommand;
import org.tranql.schema.AssociationEnd;
import org.tranql.schema.Attribute;
import org.tranql.schema.Entity;
import org.tranql.sql.Column;
import org.tranql.sql.Table;
import org.tranql.sql.jdbc.InputBinding;
import org.tranql.sql.jdbc.JDBCQueryCommand;
import org.tranql.sql.jdbc.JDBCUpdateCommand;
import org.tranql.sql.jdbc.NoOpJDBCUpdateCommand;
import org.tranql.sql.jdbc.ResultBinding;
import org.tranql.sql.jdbc.binding.BindingFactory;

/**
 * @version $Revision: 431 $ $Date: 2006-11-13 18:18:35 -0500 (Mon, 13 Nov 2006) $
 */
public class SQL92Generator implements QueryVisitor, CommandFactory {
    private final DataSource ds;

    public SQL92Generator(DataSource ds) {
        this.ds = ds;
    }

    public static class Context {
        private final Query query;
        private final StringBuffer buffer;
        private final List paramBindings = new ArrayList();
        private final Map aliasedEntity = new HashMap();
        private boolean forUpdate = false;
        private boolean aliasReferences = true;

        public Context(StringBuffer buffer, Query query) {
            this.buffer = buffer;
            this.query = query;
        }

        public StringBuffer append(Object o) {
            return buffer.append(o);
        }

        public StringBuffer append(char o) {
            return buffer.append(o);
        }

        public String getText() {
            return buffer.toString();
        }
        
        public boolean isAliasReferences() {
            return aliasReferences;
        }
        
        public void setIsAliasReferences(boolean aliasReferences) {
            this.aliasReferences = aliasReferences;
        }

        public InputBinding[] getParamBindings() {
            return (InputBinding[]) paramBindings.toArray(new InputBinding[paramBindings.size()]);
        }

		/**
		 * @return the forUpdate
		 */
		public boolean isForUpdate() {
			return forUpdate;
		}

		/**
		 * @param forUpdate the forUpdate to set
		 */
		public void setForUpdate(boolean forUpdate) {
			this.forUpdate = forUpdate;
		}
        
    }

    public QueryCommand createQuery(String sql, QueryBinding[] paramTransforms, QueryBinding[] resultTransforms) throws QueryException {
        InputBinding[] inputBindings = new InputBinding[paramTransforms.length];
        for (int i = 0; i < inputBindings.length; i++) {
            inputBindings[i] = getInputBinding(i + 1, paramTransforms[i]);
        }

        ResultBinding[] resultBindings = new ResultBinding[resultTransforms.length];
        for (int i = 0; i < resultBindings.length; i++) {
            resultBindings[i] = getResultBinding(i + 1, resultTransforms[i]);
        }

        return new JDBCQueryCommand(null, ds, sql, inputBindings, resultBindings);
    }

    public QueryCommand createQuery(Query query) throws QueryException {
        QueryBinding[] resultTransforms = query.getResultTransforms();
        ResultBinding[] resultBindings = new ResultBinding[resultTransforms.length];
        for (int i = 0; i < resultBindings.length; i++) {
            resultBindings[i] = getResultBinding(i + 1, resultTransforms[i]);
        }
        Context ctx = new Context(new StringBuffer(1024), query);
        query.visit(this, ctx);
        return new JDBCQueryCommand(query, ds, ctx.getText(), ctx.getParamBindings(), resultBindings);
    }

    public UpdateCommand createUpdate(Query query) throws QueryException {
        if ( NoOpQuery.NO_OP_QUERY == query ) {
            return new NoOpJDBCUpdateCommand();
        }
        
        QueryBinding[] resultTransforms = query.getResultTransforms();
        ResultBinding[] resultBindings = new ResultBinding[resultTransforms.length];
        for (int i = 0; i < resultBindings.length; i++) {
            resultBindings[i] = getResultBinding(i + 1, resultTransforms[i]);
        }
        Context ctx = new Context(new StringBuffer(1024), query);
        query.visit(this, ctx);
        return new JDBCUpdateCommand(null, ds, ctx.getText(), ctx.getParamBindings());
    }

    public Object visit(Query node, Object param) throws QueryException {
        Context ctx = (Context) param;
        for (Node child = node.getChild(); child != null; child = child.getSibling()) {
            param = child.visit(this, param);
        }
        if (ctx.forUpdate) {
            ctx.append(" FOR UPDATE");
        }
        return param;
    }

    public Object visit(Insert node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.aliasReferences = false;
        ctx.append("INSERT INTO ").append(node.getEntity().getPhysicalName()).append(' ');
        Node child = node.getChild();
        child.visit(this, param);
        child = child.getSibling();
        child.visit(this, param);
        return param;
    }

    public Object visit(Update node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.aliasReferences = false;
        ctx.append("UPDATE ").append(node.getEntity().getPhysicalName());
        // SET
        Node child = node.getChild();
        child.visit(this, param);

        // WHERE
        child = child.getSibling();
        if (child != null) {
            child.visit(this, param);
        }
        return param;
    }

    public Object visit(SetList node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(" SET ");
        visitChildren(node, param, ", ");
        return param;
    }

    public Object visit(Assignment node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        child.visit(this, param);
        ctx.append(" = ");
        child.getSibling().visit(this, param);
        return param;
    }

    public Object visit(ConditionalAssignment node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node set = node.getChild();
        Node flag = set.getSibling();
        Node value = flag.getSibling();
        set.visit(this, param);
        ctx.append(" = CASE WHEN ");
        flag.visit(this, param);
        ctx.append(" THEN ");
        value.visit(this, param);
        ctx.append(" ELSE ");
        set.visit(this, param);
        ctx.append(" END");
        return param;
    }

    public Object visit(Delete node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.aliasReferences = false;
        ctx.append("DELETE FROM ").append(node.getEntity().getPhysicalName());
        Node child = node.getChild();
        if (child != null) {
            child.visit(this, param);
        }
        return param;
    }

    public Object visit(Select node, Object param) throws QueryException {
        Context ctx = (Context) param;
        boolean oldAlias = ctx.aliasReferences;
        try {
            ctx.aliasReferences = true;
            Node child = node.getChild();
            if (child == null) {
                throw new MalformedQueryException("SELECT has no children");
            }
            ctx.append("SELECT ");
            if (node.isDistinct()) {
                ctx.append("DISTINCT ");
            }
            do {
                child.visit(this, param);
                child = child.getSibling();
                if (child != null) {
                    ctx.append(", ");
                }
            } while (child != null);
            return param;
        } finally {
            ctx.aliasReferences = oldAlias;
        }
    }

    public Object visit(From node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if (node.getChild() == null) {
            throw new MalformedQueryException("FROM has no child");
        }
        ctx.append(" FROM ");
        Node child = node.getChild();
        return child.visit(this, param);
    }

    public Object visit(Where node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if (node.getChild() == null) {
            throw new MalformedQueryException("WHERE has no child");
        }
        ctx.append(" WHERE ");
        Node child = node.getChild();
        return child.visit(this, param);
    }

    public Object visit(OrderBy node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if (node.getChild() == null) {
            throw new MalformedQueryException("ORDER BY has no child");
        }
        ctx.append(" ORDER BY ");
        visitChildren(node, param, ", ");
        if (node.isAsc()) {
            ctx.append(" ASC");
        } else {
            ctx.append(" DESC");
        }
        return param;
    }

    public Object visit(QuerySource node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Entity entity = node.getEntity();
        ctx.append(entity.getPhysicalName());

        String alias = node.getAlias();
        if (alias != null) {
            ctx.append(' ').append(alias);
            ctx.aliasedEntity.put(alias, entity);
        }

        if (node.isForUpdate()) {
            ctx.forUpdate = true;
        }
        return param;
    }

    public Object visit(DerivedTable node, Object param) throws QueryException {
        throw new QueryException("SQL 92 does not support the IN, collection_valued_path_expression, operator");
    }
    
    protected static final Map JOIN_TYPES = new HashMap();

    static {
        JOIN_TYPES.put(Join.CROSS_JOIN, " CROSS JOIN ");
        JOIN_TYPES.put(Join.INNER_JOIN, " INNER JOIN ");
        JOIN_TYPES.put(Join.LEFT_JOIN, " LEFT JOIN ");
        JOIN_TYPES.put(Join.RIGHT_JOIN, " RIGHT JOIN ");
        JOIN_TYPES.put(Join.FULL_JOIN, " FULL OUTER JOIN ");
    }

    public Object visit(Join node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node left = node.getChild();
        left.visit(this, param);
        ctx.append(JOIN_TYPES.get(node.getType()));
        Node right = left.getSibling();
        right.visit(this, param);
        Node condition = right.getSibling();
        if (condition != null) {
            ctx.append(" ON (");
            condition.visit(this, param);
            ctx.append(')');
        }
        return param;
    }

    public Object visit(Path node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(node.getRoot().getAlias());
        List path = node.getPath();
        for (Iterator iterator = path.iterator(); iterator.hasNext();) {
            Object o = iterator.next();
            if (o instanceof Attribute) {
                ctx.append('.').append(((Attribute) o).getPhysicalName());
            } else if (o instanceof AssociationEnd) {
                ctx.append('.').append(((AssociationEnd) o).getName());
            }
        }
        visitChildren(node, param, ".");
        return param;
    }

    public Object visit(EntityReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        AliasedEntity alias = node.getEntity();
        Table table = (Table) alias.getEntity();
        List pkFields = table.getPrimaryKeyFields();
        for (int i = 0; i < pkFields.size(); i++) {
            Column column = (Column) pkFields.get(i);
            if ( 0 < i ) {
                ctx.append(", ");
            }
            if (ctx.aliasReferences) {
                ctx.append(alias.getAlias()).append('.');
            }
            ctx.append(column.getPhysicalName());
        }
        return param;
    }

    public Object visit(AttributeReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if ( ctx.aliasReferences ) {
            ctx.append(node.getSource().getAlias()).append('.');
        }
        ctx.append(node.getAttribute().getPhysicalName());
        return param;
    }

    public Object visit(FKAttributeReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if ( ctx.aliasReferences ) {
            ctx.append(node.getSource().getAlias()).append('.');
        }
        ctx.append(node.getFKAttribute().getPhysicalName());
        return param;
    }
    
    public Object visit(TableConstructor node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(" VALUES ");
        if (node.isOnlyChild()) {
            node.getChild().visit(this, param);
        } else {
            ctx.append('(');
            visitChildren(node, param, ", ");
            ctx.append(')');
        }
        return null;
    }

    public Object visit(RowConstructor node, Object param) throws QueryException {
        generateList(node, param);
        return param;
    }

    public Object visit(Function node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(node.getFunctionType().getName());
        generateList(node, param);
        return param;
    }
    
    public Object visit(AggregateFunction node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(node.getFunction());
        ctx.append('(');
        if (node.isDistinct()) {
            ctx.append("DISTINCT ");
        }
        Node child = node.getChild();
        child.visit(this, param);
        ctx.append(')');
        return param;
    }

    public Object visit(CountAggregateFunction node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(node.getFunction());
        ctx.append('(');
        if (node.isDistinct()) {
            ctx.append("DISTINCT ");
        }
        Node child = node.getChild();
        child.visit(this, param);
        ctx.append(')');
        return param;
    }
    
    public Object visit(Not node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append("NOT (");
        node.getChild().visit(this, param);
        ctx.append(')');
        return param;
    }

    public Object visit(BinaryOperation node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node left = node.getChild();
        left.visit(this, param);
        ctx.append(' ').append(node.getOperation()).append(' ');
        Node right = left.getSibling();
        right.visit(this, param);
        return param;
    }

    public Object visit(IsNull node, Object param) throws QueryException {
        Context ctx = (Context) param;
        node.getChild().visit(this, param);
        if (node.isNot()) {
            ctx.append(" IS NOT NULL");
        } else {
            ctx.append(" IS NULL");
        }
        return param;
    }

    public Object visit(Between node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        child.visit(this, param);
        if (node.isNot()) {
            ctx.append(" NOT");
        }
        ctx.append(" BETWEEN ");
        child = child.getSibling();
        child.visit(this, param);
        ctx.append(" AND ");
        child = child.getSibling();
        child.visit(this, param);
        return param;
    }

    public Object visit(In node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        child.visit(this, param);
        if (node.isNot()) {
            ctx.append(" NOT");
        }
        ctx.append(" IN (");

        child = child.getSibling();
        child.visit(this, param);
        while ((child = child.getSibling()) != null) {
            ctx.append(", ");
            child.visit(this, param);
        }
        ctx.append(")");
        return param;
    }

    public Object visit(Like node, Object param) throws QueryException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        child.visit(this, param);
        if (node.isNot()) {
            ctx.append(" NOT");
        }
        ctx.append(" LIKE ");
        child = child.getSibling();
        child.visit(this, param);
        if (null != child.getSibling()) {
            ctx.append(" ESCAPE ");
            child = child.getSibling();
            child.visit(this, param);
        }
        return param;
    }

    public Object visit(Empty node, Object param) throws QueryException {
        throw new QueryException("SQL 92 does not support the IsEmpty operator");
    }

    public Object visit(Exists node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if (node.isNot()) {
            ctx.append("NOT ");
        }
        ctx.append("EXISTS (");
        Node select = node.getChild();
        param = select.visit(this, param);
        Node from = select.getSibling();
        param = from.visit(this, param);
        Node where = from.getSibling();
        param = where.visit(this, param);
        ctx.append(')');
        return param;
    }
    
    public Object visit(MemberOf node, Object param) throws QueryException {
        throw new QueryException("SQL 92 does not support the MemberOf operator");
    }

    public Object visit(SubQuery node, Object param) throws QueryException {
        visitChildren(node, param, "");
        return param;
    }

    public Object visit(SubQuerySource node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append('(');
        node.getSubQuery().visit(this, param);
        ctx.append(") AS ");
        ctx.append(node.getAlias());
        ctx.aliasedEntity.put(node.getAlias(), node.getEntity());
        return param;
    }

    public Object visit(StringLiteral node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append('\'');
        char[] value = node.getValue().toCharArray();
        for (int i = 0; i < value.length; i++) {
            char c = value[i];
            if (c == '\'') {
                ctx.append('\'');
            }
            ctx.append(c);
        }
        ctx.append('\'');
        return param;
    }

    public Object visit(NumericLiteral node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append(node.getValue().toString());
        return param;
    }

    public Object visit(BooleanLiteral node, Object param) throws QueryException {
        Context ctx = (Context) param;
        if (node.getValue()) {
            ctx.append("(1=1)");
        } else {
            ctx.append("(1=0)");
        }
        return param;
    }

    public Object visit(ParameterReference node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append('?');
        QueryBinding transform = ctx.query.getParamTransform(node.getParam());
        int paramIndex = ctx.paramBindings.size() + 1;
        ctx.paramBindings.add(getInputBinding(paramIndex, transform));
        return param;
    }

    public Object visit(UnaryOperator node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append('-');
        param = node.getChild().visit(this, param);
        return param;
    }
    
    protected void generateList(Node node, Object param) throws QueryException {
        Context ctx = (Context) param;
        ctx.append('(');
        visitChildren(node, param, ", ");
        ctx.append(')');
    }

    private void visitChildren(Node node, Object param, String sep) throws QueryException {
        Context ctx = (Context) param;
        Node child = node.getChild();
        if (child != null) {
            child.visit(this, param);
            while ((child = child.getSibling()) != null) {
                ctx.append(sep);
                child.visit(this, param);
            }
        }
    }

    public ResultBinding getResultBinding(int index, QueryBinding transform) throws QueryException {
        return BindingFactory.getResultBinding(index, transform);
    }

    public InputBinding getInputBinding(int index, QueryBinding transform) throws QueryException {
        return BindingFactory.getInputBinding(index, transform);
    }

}
